# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from __future__ import with_statement

import os
from types import ListType
from cStringIO import StringIO

from paste.fixture import TestApp
from lxml import etree as ET

from nagare.namespaces import xhtml
from nagare.namespaces import xhtml_base
from nagare.namespaces import xml
from nagare import component
from nagare import wsgi
from nagare import local
from nagare.sessions.memory_sessions import SessionsWithPickledStates
from nagare import presentation


def create_FixtureApp(app):
    local.worker = local.Process()
    local.request = local.Process()

    app = wsgi.create_WSGIApp(app)
    app.set_sessions_manager(SessionsWithPickledStates())
    app.start()

    return TestApp(app)


def c14n(node):
    if not isinstance(node, basestring):
        node = node.write_xmlstring()

    node = ET.fromstring(node).getroottree()

    buf = StringIO()
    node.write_c14n(buf)
    return buf.getvalue().replace('\n', '')


# Test for XHTML namespace

def head_render_init_test1():
    """ XHTML namespace unit test - HeadRender - Allowed tags """
    try:
        h = xhtml.HeadRenderer('/tmp/static_directory/')
        with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
            h << h.base({'id': 'id', 'href': 'href', 'target': 'target'})
            h << h.link({'charset': 'charset', 'href': 'href', 'hreflang': 'hreflang', 'type': 'type', 'rel': 'rel', 'rev': 'rev', 'media': 'media', 'target': 'target'})
            h << h.meta({'id': 'id', 'http_equiv': 'http_equiv', 'name': 'name', 'content': 'content', 'scheme': 'scheme'})
            h << h.title({'title': 'title'})
            h << h.style({'id': 'id'})
            h << h.script({'id': 'id'})
    except AttributeError:
        assert False
    else:
        assert True


def head_render_init_test2():
    """ XHTML namespace unit test - HeadRender - Forbiden tag """
    try:
        h = xhtml.HeadRenderer('/tmp/static_directory/')
        h << h.test({'id': 'id', 'href': 'href', 'target': 'target'})
    except AttributeError:
        assert True
    else:
        assert False


def head_render_css_url_test1():
    """ XHTML namespace unit test - HeadRender - css_url - relative url """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.css_url('css')

    assert h._css_url == {'/tmp/static_directory/css': (0, {})}


def head_render_javascript_css_test2():
    """ XHTML namespace unit test - HeadRender - css_url - absolute url """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.css_url('/css')

    assert h._css_url == {'/css': (0, {})}


def head_render_css_url_test3():
    """ XHTML namespace unit test - HeadRender - css_url - absolute url + relative url """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.css_url('/css')
        h << h.css_url('css')

    assert h._css_url['/css'] == (0, {})
    assert h._css_url['/tmp/static_directory/css'] == (1, {})


def head_render_javascript_url_test1():
    """ XHTML namespace unit test - HeadRender - javascript_url - relative url """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript_url('test.js')

    assert h._javascript_url == {'/tmp/static_directory/test.js': (0, {})}


def head_render_javascript_url_test2():
    """ XHTML namespace unit test - HeadRender - javascript_url - absolute url """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript_url('/test.js')
    assert h._javascript_url == {'/test.js': (0, {})}


def head_render_javascript_url_test3():
    """ XHTML namespace unit test - HeadRender - javascript_url - absolute url + relative url """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript_url('/test.js')
        h << h.javascript_url('test.js')

    assert h._javascript_url['/test.js'] == (0, {})
    assert h._javascript_url['/tmp/static_directory/test.js'] == (1, {})


def head_render_javascript_url_test4():
    """ XHTML namespace unit test - HeadRender - javascript_url - Add twice the same js_url"""
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript_url('test.js')
        h << h.javascript_url('test.js')

    assert h._javascript_url == {'/tmp/static_directory/test.js': (0, {})}


def head_render_javascript_test1():
    """ XHTML namespace unit test - HeadRender - javascript - use string """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript('test', 'function test(arg1) { return true }')

    assert h._named_javascript == {'test': (0, 'function test(arg1) { return true }', {})}


def head_render_javascript_test2():
    """ XHTML namespace unit test - HeadRender - javascript - use python method """

    def js_method(arg1):
        return True

    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript('test', js_method)

    assert h._javascript_url == {'/static/nagare/pyjslib.js': (0, {})}
    assert 'test' in h._named_javascript


def head_render_javascript_test3():
    """ XHTML namespace unit test - HeadRender - javascript - add 2 js with same name """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript('test', 'function test(arg1) { return true }')
        h << h.javascript('test', 'function test(arg1) { return true }')

    assert h._named_javascript == {'test': (0, 'function test(arg1) { return true }', {})}


def head_render_render_test1():
    """ XHTML namespace unit test - HeadRender - Render - render only style tag """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.style()
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head><style></style></head>')


def head_render_render_test2():
    """ XHTML namespace unit test - HeadRender - Render - render only css_url method """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.css_url('css')
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head><link href="/tmp/static_directory/css" type="text/css" rel="stylesheet"/></head>')


def head_render_render_test3():
    """ XHTML namespace unit test - HeadRender - Render - render only css method """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.css('css_test', 'test')
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head><style type="text/css">test</style></head>')


def head_render_render_test4():
    """ XHTML namespace unit test - HeadRender - Render - call render two times with css_url method"""
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.css_url('css')
    presentation.render(h, None, None, None)
    renderResult = presentation.render(h, None, None, None)
    assert not isinstance(renderResult, ListType)
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head><link href="/tmp/static_directory/css" type="text/css" rel="stylesheet"/></head>')


def head_render_render_test5():
    """ XHTML namespace unit test - HeadRender - Render - render only css method """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.css('css_test', 'test')
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head><style type="text/css">test</style></head>')


def head_render_render_test6():
    """ XHTML namespace unit test - HeadRender - Render - render only javascript_url method """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.javascript_url('test.js')
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head><script src="/tmp/static_directory/test.js" type="text/javascript"></script></head>')


def head_render_render_test7():
    """ XHTML namespace unit test - HeadRender - Render - render only string js method """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.javascript('test.js', 'function test() { return True }')
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head><script type="text/javascript">function test() { return True }</script></head>')


def head_render_render_test8():
    """ XHTML namespace unit test - HeadRender - Render - render only python2js method """

    def js_method(arg1):
        return True

    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'lang': 'lang', 'dir': 'dir', 'id': 'id', 'profile': 'profile'}):
        h << h.javascript('test', js_method)
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head lang="lang" profile="profile" id="id" dir="dir"><script src="/static/nagare/pyjslib.js" type="text/javascript"></script><script type="text/javascript">function nagare_namespaces_test_test_xhtmlns_js_method(arg1) {    return true;}</script></head>')


def head_render_render_test9():
    """ XHTML namespace unit test - HeadRender - Render - render with head """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    h << h.head({'id': 'id'})
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head id="id"></head>')


def head_render_render_test10():
    """ XHTML namespace unit test - HeadRender - Render - render with head & style """
    h = xhtml.HeadRenderer('/tmp/static_directory/')
    with h.head({'id': 'id'}):
        h << h.style('test', {'id': 'id'})
    assert c14n(presentation.render(h, None, None, None)) == c14n('<head id="id"><style id="id">test</style></head>')


def html_render_init_test1():
    """ XHTML namespace unit test - HTMLRender - init - test if head exists """
    h = xhtml.Renderer()
    assert hasattr(h, 'head')
    assert isinstance(h.head, xhtml.HeadRenderer)


def html_render_init_test2():
    """ XHTML namespace unit test - HTMLRender - init - add bad element """
    h = xhtml.Renderer()
    try:
        h << h.bad_element
    except AttributeError:
        assert True
    else:
        assert False


def html_render_parse_html_test1():
    """ XHTML namespace unit test - HTMLRender - parse_html - good encoding """
    try:
        h = xhtml.Renderer()
        h.parse_html(os.path.join(os.path.dirname(__file__), 'helloworld.html'))
    except UnicodeDecodeError:
        assert False
    else:
        assert True


if 0:
    def html_render_parse_html_test2():
        """ XHTML namespace unit test - HTMLRender - parse_html - bad encoding """
        try:
            h = xhtml.Renderer()
            h.parse_html(os.path.join(os.path.dirname(__file__), 'iso-8859.html'), encoding='utf-8')
        except UnicodeDecodeError:
            assert True
        else:
            assert False


def html_render_parse_html_test3():
    """ XHTML namespace unit test - HTMLRender - parse_htmlstring - bad html (auto correct)"""
    h = xhtml.Renderer()
    root = h.parse_htmlstring('<html><head><body></body></head><html>')
    assert c14n(root) == c14n('<html><head></head><body></body></html>')


def html_render_parse_html_test4():
    """ XHTML namespace unit test - HTMLRender - parse_htmlstring - bad html"""
    h = xhtml.Renderer()
    root = h.parse_htmlstring('test')
    assert c14n(root) == c14n('<html><body><p>test</p></body></html>')


def html_render_parse_html_test5():
    """ XHTML namespace unit test - HTMLRender - parse_html - get html from an URL """
    try:
        h = xhtml.Renderer()
        h.parse_html('http://www.google.fr/')
    except:
        assert False
    else:
        assert True


xml_tree_1 = "<a>text</a>"


def html_render_parse_html_test6():
    """ XHTML namespace unit test - HTMLRender - parse_html - option xhtml=False - xmlstring - Test child type """
    x = xhtml.Renderer()
    root = x.parse_htmlstring(xml_tree_1, xhtml=False)
    assert type(root) == xhtml_base._HTMLTag


def html_render_parse_html_test7():
    """ XHTML namespace unit test - HTMLRender - parse_html - option xhtml=True  - xmlstring - Test child type """
    x = xhtml.Renderer()
    root = x.parse_htmlstring(xml_tree_1, xhtml=True)
    assert type(root) == xhtml.A


def html_render_parse_html_test8():
    """ XHTML namespace unit test - HTMLRender - parse_xml - xmlstring - Test child type """
    x = xhtml.Renderer()
    root = x.parse_xmlstring(xml_tree_1)
    assert type(root) == xml._Tag


html_tree_1 = "<html><body></html>"


def html_render_parse_html_test9():
    """ XHTML namespace unit test - HTMLRender - parse_html - option xhtml=False - htmlstring - Test child type """
    x = xhtml.Renderer()
    root = x.parse_htmlstring(html_tree_1, xhtml=False)
    assert type(root) == xhtml_base._HTMLTag


def html_render_parse_html_test10():
    """ XHTML namespace unit test - HTMLRender - parse_html - option xhtml=True  - htmlstring - Test child type """
    try:
        x = xhtml.Renderer()
        x.parse_htmlstring(html_tree_1, xhtml=True)
    except:
        assert True
    else:
        assert False


def html_render_parse_html_test11():
    """ XHTML namespace unit test - HTMLRender - parse_xml - htmlstring - Test child type """
    try:
        x = xhtml.Renderer()
        x.parse_xmlstring(html_tree_1)
    except:
        assert True
    else:
        assert False


xhtml_tree_1 = "<xhtml><a/></xhtml>"


def html_render_parse_html_test12():
    """ XHTML namespace unit test - HTMLRender - parse_html - option xhtml=False - xhtmlstring - Test child type """
    x = xhtml.Renderer()
    root = x.parse_htmlstring(xhtml_tree_1, xhtml=False)
    assert type(root.getchildren()[0]) == xhtml_base._HTMLTag


def html_render_parse_html_test13():
    """ XHTML namespace unit test - HTMLRender - parse_html - option xhtml=True - xhtmlstring - Test child type """
    x = xhtml.Renderer()
    root = x.parse_htmlstring(xhtml_tree_1, xhtml=True)
    assert type(root.getchildren()[0]) == xhtml.A


def html_render_parse_html_test14():
    """ XHTML namespace unit test - HTMLRender - parse_xml - xhtmlstring - Test child type """
    x = xhtml.Renderer()
    root = x.parse_xmlstring(xhtml_tree_1)
    assert type(root.getchildren()[0]) == xml._Tag


def htmltag_write_xmlstring_test1():
    """ XHTML namespace unit test - HTMLRender - write_htmlstring - without argument """
    h = xhtml.Renderer()
    h << h.table(h.tr(h.td()), h.tr(h.td()))
    assert c14n(h.root) == c14n('<table><tr><td></td></tr><tr><td></td></tr></table>')


def htmltag_write_xmlstring_test2():
    """ XHTML namespace unit test - HTMLRender - write_htmlstring - with pipeline == True"""
    h = xhtml.Renderer()
    h << h.table(h.tr(h.td().meld_id('test'), h.tr(h.td().meld_id('test'))))
    assert c14n(h.root.write_htmlstring(pipeline=True)) == c14n('<table><tr><td xmlns:ns0="http://www.plope.com/software/meld3" ns0:id="test"></td><tr><td xmlns:ns0="http://www.plope.com/software/meld3" ns0:id="test"></td></tr></tr></table>')


def htmltag_write_xmlstring_test3():
    """ XHTML namespace unit test - HTMLRender - write_htmlstring - with pipeline == False """
    h = xhtml.Renderer()
    h << h.table(h.tr(h.td().meld_id('false'), h.tr(h.td().meld_id('false'))))
    assert c14n(h.root.write_htmlstring(pipeline=False)) == c14n('<table><tr><td xmlns:ns0="http://www.plope.com/software/meld3"></td><tr><td xmlns:ns0="http://www.plope.com/software/meld3"></td></tr></tr></table>')


def html_render_add_tag_test1():
    """ XHTML namespace unit test - HTMLRender - add tag - create simple html """
    h = xhtml.Renderer()
    h << h.html(h.body(h.table(h.tr(h.td()), h.tr(h.td()))))
    assert c14n(h.root) == c14n('<html><body><table><tr><td></td></tr><tr><td></td></tr></table></body></html>')


def html_render_form_test1():
    """ XHTML namespace unit test - Form - create simple form """
    h = xhtml.Renderer()
    h << h.html(h.body(h.form(h.input(type="string", name="input1", value="value"), h.input(type="submit", name="submit"))))
    form = h.root.xpath('.//form')
    attributes = form[0].attrib
    assert attributes['method'] == "post"
    assert attributes['accept-charset'] == "utf-8"
    assert attributes['enctype'] == "multipart/form-data"
    assert attributes['action'] == "?"

    assert c14n(h.root) == c14n('<html><body><form enctype="multipart/form-data" method="post" accept-charset="utf-8" action="?"><input type="string" name="input1" value="value"/><input type="submit" name="submit"/></form></body></html>')


def html_render_form_test2():
    """ XHTML namespace unit test - Form - create 2 imbrecated forms """
    h = xhtml.Renderer()
    h << h.html(h.body(h.form(h.input(type="string", name="input1", value="value"), h.form(h.input(type="submit", name="submit")))))
    assert len(h.root.xpath('.//form')) == 1


def html_render_form_test3():
    """ XHTML namespace unit test - Form - create 2 forms not imbrecated"""
    h = xhtml.Renderer()
    h << h.html(h.body(h.form(h.input(type="string", name="input1", value="value")), h.form(h.input(type="string", name="input2", value="value"))))
    assert len(h.root.xpath('.//form')) == 2


def html_render_form_test4():
    """ XHTML namespace unit test - Form - create simple form with iso-8859-15 enconding"""
    h = xhtml.Renderer()
    h << h.html(h.body(h.form(h.input(type="string", name="input1", value="value"), h.input(type="submit", name="submit"), {"accept-charset": "iso-8859-15"})))
    form = h.root.xpath('.//form')
    attributes = form[0].attrib
    assert attributes['method'] == "post"
    assert attributes['accept-charset'] == "iso-8859-15"
    assert attributes['enctype'] == "multipart/form-data"
    assert attributes['action'] == "?"


class My_app_test_form():

    def __init__(self):
        self.actions_done = []

    def my_pre_action(self):
        self.actions_done.append('my_pre_action')
        assert self.actions_done == ['my_pre_action']

    def my_post_action(self):
        self.actions_done.append('my_post_action')
        assert self.actions_done == ['my_pre_action', 'my_input_action', 'my_post_action']

    def my_input_action(self, text):
        self.actions_done.append('my_input_action')
        assert self.actions_done == ['my_pre_action', 'my_input_action']

    def my_submit_action(self):
        self.actions_done.append('my_submit_action')
        assert self.actions_done == ['my_pre_action', 'my_input_action', 'my_post_action', 'my_submit_action']


@presentation.render_for(My_app_test_form)
def render(self, h, *args):
    h << h.html(h.body(h.div('My_app_test_form'), h.form(
        h.input(type="text", name="input1", value="value").action(self.my_input_action),
        h.input(type="submit", name="submit").action(self.my_submit_action)
    ).pre_action(self.my_pre_action).post_action(self.my_post_action)))
    return h.root


def html_render_form_test5():
    """ XHTML namespace unit test - Form - create test methods call order """
    myApp = My_app_test_form
    app = create_FixtureApp(myApp)
    res = app.get('/')
    form = res.forms[0]
    res = form.submit()


def html_render_select_test1():
    """ XHTML namespace unit test - Select - form with simple select input """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                with h.select:
                    h << h.option(value="option1")
                    h << h.option(value="option2", selected='selected')
    options = h.root.xpath('.//option')
    assert options[1].attrib['selected'] == 'selected'


def html_render_select_test2():
    """ XHTML namespace unit test - Select - test selected method """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                with h.select:
                    h << h.option(value="option1")
                    h << h.option(value="option2")

    options = h.root.xpath('.//option')
    for option in options:
        option.selected('option2')

    assert options[1].attrib['selected'] == 'selected'


def html_render_select_test3():
    """ XHTML namespace unit test - Select - test selected method with multiple select """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                with h.select(multiple=True):
                    h << h.option(value="option1")
                    h << h.option(value="option2")
                    h << h.option(value="option3")

    options = h.root.xpath('.//option')
    for option in options:
        option.selected(['option1', 'option3'])

    assert options[0].attrib['selected'] == 'selected'
    assert options[2].attrib['selected'] == 'selected'


def html_render_select_test4():
    """ XHTML namespace unit test - Select - test selected method with multiple select and prefixed select """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                with h.select(multiple=True):
                    h << h.option(value="option1")
                    h << h.option(value="option2", selected="selected")
                    h << h.option(value="option3")

    options = h.root.xpath('.//option')
    for option in options:
        option.selected(['option1', 'option3'])

    assert options[0].attrib['selected'] == 'selected'
    assert options[2].attrib['selected'] == 'selected'


def html_render_select_test5():
    """ XHTML namespace unit test - Select - test selected method with multiple select and option group """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                with h.select(multiple=True):
                    h << h.option(value="option1")
                    with h.optgroup:
                        h << h.option(value="option2")
                        h << h.option(value="option3")

    options = h.root.xpath('.//option')
    for option in options:
        option.selected(['option1', 'option3'])

    assert options[0].attrib['selected'] == 'selected'
    assert options[2].attrib['selected'] == 'selected'


def html_render_select_test6():
    """ XHTML namespace unit test - Select - test selected method with multiple select """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                h << h.select([h.option(value='option1').selected(['option1', 'option3']),
                               h.option(value='option2').selected(['option1', 'option3']),
                               h.option(value='option3').selected(['option1', 'option3'])], multiple=True)

    options = h.root.xpath('.//option')
    assert options[0].attrib['selected'] == 'selected'
    assert options[2].attrib['selected'] == 'selected'


class My_app_test_select_multiple():

    def __init__(self):
        self.choices = "default"

    def set_choices(self, choices):
        assert isinstance(choices, (list, tuple))
        self.choices = choices


@presentation.render_for(My_app_test_select_multiple)
def render_2(self, h, *args):
    with h.html:
        with h.body:
            h << h.div("choice:", self.choices)
            with h.form:
                h << h.select([h.option(value='option1').selected(['option2']),
                               h.option(value='option2').selected(['option2']),
                               h.option(value='option3').selected(['option2'])], multiple=True).action(self.set_choices)
    return h.root


def html_render_select_test7():
    """ XHTML namespace unit test - Select - test selected method with multiple attribute but one choice """
    myApp = My_app_test_select_multiple
    app = create_FixtureApp(myApp)
    res = app.get('/')
    assert 'choice:default' in res
    form = res.forms[0]
    res = form.submit()
    assert 'choice:option2' in res


class My_app_test_select_single():

    def __init__(self):
        self.choices = "default"

    def set_choice(self, choices):
        assert choices == 'option1'
        self.choices = choices


@presentation.render_for(My_app_test_select_single)
def render_3(self, h, *args):
    with h.html:
        with h.body:
            h << h.div("choice:", self.choices)
            with h.form:
                h << h.select([h.option(value="option1").selected(['option1']),
                               h.option(value="option2").selected(['option1']),
                               h.option(value="option3").selected(['option1'])]).action(self.set_choice)
    return h.root


def html_render_select_test8():
    """ XHTML namespace unit test - Select - test selected method with single attribute """
    myApp = My_app_test_select_single
    app = create_FixtureApp(myApp)
    res = app.get('/')
    assert 'choice:default' in res
    form = res.forms[0]
    res = form.submit()
    assert 'choice:option1' in res


def html_render_checkbox_test1():
    """ XHTML namespace unit test - Checkboxes - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                h << h.input(type="checkbox", value="option1")
                h << h.input(type="checkbox", value="option2", checked='checked')
    checkboxes = h.root.xpath('.//input[@type="checkbox"]')
    assert 'checked' not in checkboxes[0].attrib.keys()
    assert 'checked' in checkboxes[1].attrib.keys()


def html_render_checkbox_test2():
    """ XHTML namespace unit test - Checkboxes - test selected method """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                checkbox = h.input(type="checkbox", value="option1")
                h << checkbox
                h << h.input(type="checkbox", value="option2", checked='checked')
    checkboxes = h.root.xpath('.//input[@type="checkbox"]')
    checkbox.selected(True)

    assert 'checked' in checkboxes[0].attrib.keys()
    assert 'checked' in checkboxes[1].attrib.keys()


def html_render_checkbox_test3():
    """ XHTML namespace unit test - Checkboxes - test selected method to unselected checkbox """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                checkbox = h.input(type="checkbox", value="option1", checked='checked')
                h << checkbox
                h << h.input(type="checkbox", value="option2")
    checkboxes = h.root.xpath('.//input[@type="checkbox"]')
    checkbox.selected(False)

    assert 'checked' not in checkboxes[0].attrib.keys()


def html_render_radiobutton_test1():
    """ XHTML namespace unit test - Radio - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                h << h.input(name="radio", type="radio", value="option1")
                h << h.input(name="radio", type="radio", value="option2", checked='checked')
                h << h.input(name="radio", type="radio", value="option3")
    radios = h.root.xpath('.//input[@type="radio"]')
    assert 'checked' not in radios[0].attrib.keys()
    assert 'checked' in radios[1].attrib.keys()
    assert 'checked' not in radios[2].attrib.keys()


def html_render_radiobutton_test2():
    """ XHTML namespace unit test - Radio - test selected method """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                radio = h.input(name="radio", type="radio", value="option1")
                h << radio
                h << h.input(name="radio", type="radio", value="option2", checked='checked')
                h << h.input(name="radio", type="radio", value="option3")
    radios = h.root.xpath('.//input[@type="radio"]')
    radio.selected(True)
    assert 'checked' in radios[0].attrib.keys()
    assert 'checked' in radios[1].attrib.keys()
    assert 'checked' not in radios[2].attrib.keys()


def html_render_radiobutton_test3():
    """ XHTML namespace unit test - Radio - test selected method to unselected checkbox """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                h << h.input(name="radio", type="radio", value="option1")
                radio = h.input(name="radio", type="radio", value="option2", checked='checked')
                h << radio
                h << h.input(name="radio", type="radio", value="option3")
    radios = h.root.xpath('.//input[@type="radio"]')
    radio.selected(False)
    assert 'checked' not in radios[0].attrib.keys()
    assert 'checked' not in radios[1].attrib.keys()
    assert 'checked' not in radios[2].attrib.keys()


def html_render_radiobutton_test4():
    """ XHTML namespace unit test - Radio - init with label """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                h << h.input(name="radio", type="radio", value="value1", id="option1") << h.label("option 1", for_="option1") << h.br
                h << h.input(name="radio", type="radio", value="value1", id="option2", checked="checked") << h.label("option 2", for_="option2") << h.br
                h << h.input(name="radio", type="radio", value="value1", id="option3") << h.label("option 3", for_="option3") << h.br
    labels = h.root.xpath('.//label')

    assert reduce(lambda x, y: x and y, [label.get('for') is not None for label in labels])


def html_render_textarea_test1():
    """ XHTML namespace unit test - TextArea - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                h << h.textarea("test", name="textearea1", type="textarea")

    textarea = h.root.xpath('.//textarea')[0]
    assert textarea.text == 'test'


def html_render_submit_test1():
    """ XHTML namespace unit test - Input submit - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                input = h.input(name="submit1", type="submit")
                h << input

    assert isinstance(input, xhtml.SubmitInput)


def html_render_password_test1():
    """ XHTML namespace unit test - Input password - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                input = h.input(name="password1", type="password")
                h << input

    assert isinstance(input, xhtml.PasswordInput)


def html_render_hidden_test1():
    """ XHTML namespace unit test - Input hidden - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                input = h.input(name="hidden1", type="hidden")
                h << input

    assert isinstance(input, xhtml.HiddenInput)


def html_render_file_test1():
    """ XHTML namespace unit test - Input file - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                input = h.input(name="file1", type="file")
                h << input

    assert isinstance(input, xhtml.FileInput)


def html_render_text_test1():
    """ XHTML namespace unit test - Input text - init """
    h = xhtml.Renderer()
    with h.html:
        with h.body:
            with h.form:
                input = h.input(name="text1", type="text")
                h << input

    assert isinstance(input, xhtml.TextInput)


def html_render_img_test1():
    """ XHTML namespace unit test - Tag img - init - external source """
    h = xhtml.Renderer()

    with h.html:
        with h.body:
            h << h.img(src="http://www.google.com/intl/en_ALL/images/logo.gif")

    assert c14n(h.root) == c14n('<html><body><img src="http://www.google.com/intl/en_ALL/images/logo.gif"/></body></html>')


def html_render_img_test2():
    """ XHTML namespace unit test - Tag img - init - relative source """
    h = xhtml.Renderer(static_url='/tmp/static/')

    with h.html:
        with h.body:
            h << h.img(src="logo.gif")

    assert c14n(h.root) == c14n('<html><body><img src="/tmp/static/logo.gif"/></body></html>')


def html_render_img_test3():
    """ XHTML namespace unit test - Tag img - init - absolute source """
    h = xhtml.Renderer(static_url='/tmp/static/')

    with h.html:
        with h.body:
            h << h.img(src="/logo.gif")

    assert c14n(h.root) == c14n('<html><body><img src="/logo.gif"/></body></html>')


def html_render_a_test1():
    """ XHTML namespace unit test - Tag a - init """
    h = xhtml.Renderer()

    with h.html:
        with h.body:
            h << h.a("google", href="http://www.google.com")

    a = h.root.xpath('.//a')[0]
    assert isinstance(a, xhtml.A)
    assert c14n(h.root) == c14n('<html><body><a href="http://www.google.com">google</a></body></html>')


def html_render_action_test1():
    """ XHTML namespace unit test - action - put action method on tag a """
    h = xhtml.Renderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.a().action(lambda x: None)
            h << a

    assert a.attrib['href'] is not None


def html_render_action_test2():
    """ XHTML namespace unit test - action - put action method on tag a (replace existing href) """
    h = xhtml.Renderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.a(href='http://www.google.com').action(lambda x: None)
            h << a

    assert a.attrib['href'] == 'http://www.google.com'


def html_render_action_test3():
    """ XHTML namespace unit test - action - put action method on tag imagge """
    h = xhtml.Renderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.img().action(lambda x: None)
            h << a

    assert a.attrib['src'] is not None


def html_render_action_test4():
    """ XHTML namespace unit test - action - put action method on tag a (replace existing href) """
    h = xhtml.Renderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.a(src="logo.gif").action(lambda x: None)
            h << a

    assert a.attrib['src'] != 'http://www.google.com'


def html_render_action_test5():
    """ XHTML namespace unit test - asynchronous render - action - put action method on tag a """
    h = xhtml.AsyncRenderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.a().action(lambda x: None)
            h << a

    assert a.attrib['href'] is not None


def html_render_action_test6():
    """ XHTML namespace unit test - asynchronous render - action - put action method on tag a (replace existing href) """
    h = xhtml.AsyncRenderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.a(href='http://www.google.com').action(lambda x: None)
            h << a

    assert a.attrib['href'] != 'http://www.google.com'


def html_render_action_test7():
    """ XHTML namespace unit test - asynchronous render - action - put action method on tag imagge """
    h = xhtml.AsyncRenderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.img().action(lambda x: None)
            h << a

    assert a.attrib['src'] is not None


def html_render_action_test8():
    """ XHTML namespace unit test - asynchronous render - action - put action method on tag a (replace existing href) """
    h = xhtml.AsyncRenderer(static_url='/tmp/static/')
    h.start_rendering(component.Component(None), None)

    with h.html:
        with h.body:
            a = h.a(src="logo.gif").action(lambda x: None)
            h << a

    assert a.attrib['src'] != 'http://www.google.com'


xml_test1_out = c14n("""<html><body onload="javascript:alert()">
<ul>
<li>Hello</li>
<li>world</li>
<li>yeah</li>
</ul>
<div class="toto"><h1>moi<i>toto</i>
</h1></div>
<div>yeah012</div>
<table toto="toto">
<tr>
<td>1</td>
<td>a</td>
</tr>
<tr>
<td>2</td>
<td>b</td>
</tr>
<tr>
<td>3</td>
<td>c</td>
</tr>
</table>
</body></html>""")


def global_test1():
    """ XHTML namespace unit test - create xhtml by procedural way """
    t = ((1, 'a'), (2, 'b'), (3, 'c'))

    h = xhtml.Renderer()

    with h.html:
        with h.body(onload='javascript:alert()'):
            with h.ul:
                with h.li('Hello'):
                    pass
                with h.li:
                    h << 'world'
                h << h.li('yeah')

            with h.div(class_='toto'):
                with h.h1('moi'):
                    h << h.i('toto')

            with h.div:
                h << 'yeah'
                for i in range(3):
                    h << i

            with h.table(toto='toto'):
                for row in t:
                    with h.tr:
                        for column in row:
                            with h.td:
                                h << column

    assert c14n(h.root) == xml_test1_out


def global_test2():
    """ XHTML namespace unit test - create xhtml by functionnal way """
    t = ((1, 'a'), (2, 'b'), (3, 'c'))

    h = xhtml.Renderer()

    helloWorld = h.ul([h.li('Hello'), h.li('world'), h.li('yeah')])
    totoDiv = h.div([h.h1('moi', [h.i('toto')])], {'class': 'toto'})
    yeah012 = h.div('yeah012')
    table = h.table([h.tr([h.td(elt1), h.td(elt2)]) for elt1, elt2 in t], {'toto': 'toto'})

    html = h.html([h.body([helloWorld, totoDiv, yeah012, table], {'onload': 'javascript:alert()'})])

    assert c14n(html) == xml_test1_out



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The XHTML5 renderer
"""

from __future__ import with_statement

from lxml import etree as ET

from nagare.namespaces.xml import TagProp
from nagare.namespaces import xhtml_base, xhtml


# -----------------------------------------------------------------------------

class Input(xhtml._HTMLActionTag):
    """Class of all the new input types of HTML5
    """
    _actions = (1, 'name', 'onchange')
    async_action = xhtml._HTMLActionTag.sync_action


class ObsoleteTagProp(TagProp):
    """Class of all the HTML4 tags obsolete in HTML5
    """
    def __get__(self, renderer, cls):
        # Raise the same exception as when an attribute doesn't exist
        raise AttributeError("'%s' object has no attribute '%s'" % (renderer.__class__.__name__, self._name))


# -----------------------------------------------------------------------------

class Source(xhtml_base._HTMLTag):
    """The <source> tag
    """
    def add_child(self, child):
        """Add child and attributes
        """
        super(Source, self).add_child(child)

        # If this is a relative URL, it's relative to the statics directory
        # of the application
        src = self.get('src')
        if src is not None:
            self.set('src', xhtml.absolute_url(src, self.renderer.head.static_url))


# -----------------------------------------------------------------------------

class Renderer(xhtml.Renderer):
    """The XHTML5 synchronous renderer
    """

    XML_DOCTYPE = HTML_DOCTYPE = '<!DOCTYPE html>'

    # New HTML5 tags
    # --------------

    section = TagProp('section')
    article = TagProp('article')
    aside = TagProp('aside')
    hgroup = TagProp('hgroup')
    header = TagProp('header')
    footer = TagProp('footer')
    nav = TagProp('nav')
    figure = TagProp('figure')
    figcaption = TagProp('figcaption')
    main = TagProp('main')
    time = TagProp('time')
    video = TagProp('video')
    audio = TagProp('audio')
    source = TagProp('source', factory=Source)
    embed = TagProp('embed')
    mark = TagProp('mark')
    meta = TagProp('meta')
    progress = TagProp('progress')
    meter = TagProp('meter')
    ruby = TagProp('ruby')
    rt = TagProp('rt')
    rp = TagProp('rp')
    wbr = TagProp('wbr')
    canvas = TagProp('canvas')
    command = TagProp('command')
    details = TagProp('details')
    summary = TagProp('summary')
    datalist = TagProp('datalist')
    keygen = TagProp('keygen')
    output = TagProp('output')
    track = TagProp('track')

    # Obsolete HTML4 tags
    # -------------------

    basefont = ObsoleteTagProp('basefont')
    big = ObsoleteTagProp('big')
    center = ObsoleteTagProp('center')
    font = ObsoleteTagProp('font')
    strike = ObsoleteTagProp('strike')
    tt = ObsoleteTagProp('tt')
    frame = ObsoleteTagProp('frame')
    frameset = ObsoleteTagProp('frameset')
    noframes = ObsoleteTagProp('u')
    acronym = ObsoleteTagProp('acronym')
    applet = ObsoleteTagProp('applet')
    isindex = ObsoleteTagProp('isindex')
    dir = ObsoleteTagProp('dir')

    # New input types
    # ---------------

    _specialTags = dict([(input_type + '_input', Input) for input_type in (
        'tel', 'search', 'url', 'email',
        'datetime', 'date', 'datetime-local_input',
        'month', 'week', 'time', 'number', 'range', 'color'
    )])
    _specialTags.update(xhtml.Renderer._specialTags)

    @classmethod
    def class_init(cls, specialTags):
        """Class initialisation

        In:
          -- ``special_tags`` -- tags that have a special factory
        """
        class CustomLookup(ET.CustomElementClassLookup):
            def __init__(self, specialTags, defaultLookup):
                super(CustomLookup, self).__init__(defaultLookup)
                self._specialTags = specialTags

            def lookup(self, node_type, document, namespace, name):
                return self._specialTags.get(name)

        cls._specialTags.update(specialTags)

        cls._custom_lookup = CustomLookup(cls._specialTags, ET.ElementDefaultClassLookup(element=xhtml_base._HTMLTag))
        cls._html_parser = ET.HTMLParser()
        cls._html_parser.setElementClassLookup(cls._custom_lookup)

    def AsyncRenderer(self, *args, **kw):
        """Create an associated asynchronous HTML renderer

        Return:
          - a new asynchronous renderer
        """
        # If no arguments are given, this renderer becomes the parent of the
        # newly created renderer
        if not args and not kw:
            args = (self,)

        return AsyncRenderer(*args, **kw)


# -----------------------------------------------------------------------------

class AsyncRenderer(xhtml.AsyncRenderer, Renderer):
    """The XHTML5 asynchronous renderer
    """
    def SyncRenderer(self, *args, **kw):
        """Create an associated synchronous HTML renderer

        Return:
          - a new synchronous renderer
        """
        # If no arguments are given, this renderer becomes the parent of the
        # newly created renderer
        if not args and not kw:
            args = (self,)

        return Renderer(*args, **kw)


# ---------------------------------------------------------------------------

if __name__ == '__main__':
    from nagare import presentation

    t = ((1, 'a'), (2, 'b'), (3, 'c'))

    h = Renderer()

    h.head << h.head.title('A test')
    h.head << h.head.javascript('__foo__', 'function() {}')
    h.head << h.head.meta(name='meta1', content='content1')

    with h.body(onload='javascript:alert()'):
        h << h.section(name='name')
        h << h.article
        h << h.aside
        h << h.hgroup
        h << h.header
        h << h.footer
        h << h.nav
        h << h.figure

        h << h.video
        h << h.audio
        h << h.source
        h << h.embed
        h << h.mark
        h << h.progress
        h << h.meter
        h << h.ruby
        h << h.rt
        h << h.rp
        h << h.wbr
        h << h.canvas
        h << h.command
        h << h.details
        h << h.summary
        h << h.datalist
        h << h.keygen
        h << h.output

        with h.ul:
            with h.li('Hello'):
                pass
            with h.li:
                h << 'world'
            h << h.li('yeah')

        with h.div(class_='foo'):
            with h.h1('moi'):
                h << h.i('foo')

        with h.div:
            h << 'yeah'
            for i in range(3):
                h << i

        with h.table(foo='foo'):
            for row in t:
                with h.tr:
                    for column in row:
                        with h.td:
                            h << column

        with h.form:
            h << h.input(type='url')

    print h.html(presentation.render(h.head, None, None, None), h.root).write_htmlstring(pretty_print=True)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The XHTML renderer

This renderer only depends of the ``nagare.namespaces.xml`` module.
Having not dependencies to the framework make it suitable to be used in others frameworks.
"""

from __future__ import with_statement

import cStringIO
import urllib
import peak

from lxml import etree as ET

from nagare.namespaces import xml
from nagare.namespaces.xml import TagProp

# ---------------------------------------------------------------------------

# Common attributes
# -----------------

componentattrs = ('id', 'class', 'style', 'title')
i18nattrs = ('lang', 'dir')
eventattrs = (
    'onclick', 'ondblclick', 'onmousedown', 'onmouseup', 'onmousemove',
    'onmouseover', 'onmouseout', 'onkeypress', 'onkeydown', 'onkeyup'
)
allattrs = componentattrs + i18nattrs + eventattrs
focusattrs = ('accesskey', 'tabindex', 'onfocus', 'onblur')


class _HTMLTag(xml._Tag):
    """A xhtml tag
    """
    def write_xmlstring(self, encoding='utf-8', pipeline=True, **kw):
        """Serialize in XML the tree beginning at this tag

        In:
          - ``encoding`` -- encoding of the XML
          - ``pipeline`` -- if False, the ``meld:id`` attributes are deleted

        Return:
          - the XML
        """
        return xml._Tag.write_xmlstring(self.decorate_error(), encoding, pipeline, **kw)

    def write_htmlstring(self, encoding='utf-8', pipeline=True, **kw):
        """Serialize in XHTML the tree beginning at this tag

        In:
          - ``encoding`` -- encoding of the XML
          - ``pipeline`` -- if False, the ``meld:id`` attributes are deleted

        Return:
          - the XHTML
        """
        if not pipeline:
            for element in self.xpath('.//*[@meld:id]', namespaces={'meld': xml._MELD_NS}):
                del element.attrib[xml._MELD_ID]

        return ET.tostring(self.decorate_error(), encoding=encoding, method='html', **kw)

    def error(self, err):
        """Mark this tag as erroneous

        In:
          - ``err`` -- the error message

        Return:
          - ``self``
        """
        if err is not None:
            self.has_error = err

        return self

    def decorate_error(self):
        """Decorate this tag with an error structure

        Forward the call to the renderer ``decorate_error()`` method

        Return:
          - a tree
        """
        err = getattr(self, 'has_error', None)
        return self if err is None else self.renderer.decorate_error(self, err)


@peak.rules.when(xml.add_child, (_HTMLTag, _HTMLTag))
def add_child(next_method, self, element):
    """Add a tag to a tag

    In:
      - ``self`` -- the tag
      - ``element`` -- the tag to add
    """
    return next_method(self, element.decorate_error())


@peak.rules.when(xml.add_attribute, (_HTMLTag, basestring, basestring))
def add_attribute(next_method, self, name, value):
    if name.startswith('data_'):
        name = name.replace('_', '-')

    next_method(self, name, value)


class HeadRenderer(xml.XmlRenderer):
    """The XHTML head Renderer

    This renderer knows about the possible tags of a html ``<head>``
    """

    # Tag factories
    # -------------

    base = TagProp('base', set(('id', 'href', 'target')))
    head = TagProp('head', set(i18nattrs + ('id', 'profile')))
    link = TagProp('link', set(allattrs + ('charset', 'href', 'hreflang', 'type', 'rel', 'rev', 'media', 'target')))
    meta = TagProp('meta', set(i18nattrs + ('id', 'http_equiv', 'name', 'content', 'scheme')))
    title = TagProp('title', set(i18nattrs + ('id',)))
    style = TagProp('style', set(i18nattrs + ('id',)))
    script = TagProp('script', set(i18nattrs + ('id',)))

    @classmethod
    def class_init(cls, specialTags):
        """Class initialisation

        In:
          -- ``special_tags`` -- tags that have a special factory
        """
        # Create a XML parser that generate ``_HTMLTag`` nodes
        cls._xml_parser = ET.XMLParser()
        cls._xml_parser.setElementClassLookup(ET.ElementDefaultClassLookup(element=_HTMLTag))


class Renderer(xml.XmlRenderer):
    head_renderer_factory = HeadRenderer

    componentattrs = ('id', 'class', 'style', 'title')
    i18nattrs = ('lang', 'dir')
    eventattrs = (
        'onclick', 'ondblclick', 'onmousedown', 'onmouseup', 'onmousemove',
        'onmouseover', 'onmouseout', 'onkeypress', 'onkeydown', 'onkeyup'
    )
    focusattrs = ('accesskey', 'tabindex', 'onfocus', 'onblur')
    cellhalignattrs = ('align', 'char', 'charoff')
    cellvalignattrs = ('valign',)
    allattrs = componentattrs + i18nattrs + eventattrs

    # The HTML tags
    # -------------

    a = TagProp('a', set(allattrs + focusattrs + (
        'charset', 'type', 'name', 'href', 'hreflang', 'rel',
        'rev', 'shape', 'coords', 'target', 'oncontextmenu'
    )))
    abbr = TagProp('abbr', set(allattrs))
    acronym = TagProp('acronym', set(allattrs))
    address = TagProp('address', set(allattrs))
    applet = TagProp('applet', set(componentattrs + (
        'codebase', 'archive', 'code', 'object', 'alt', 'name', 'width',
        'height', 'align', 'hspace', 'vspace'
    )))
    area = TagProp('area', set(allattrs + focusattrs + ('shape', 'coords', 'href', 'nohref', 'alt', 'target')))
    b = TagProp('b', set(allattrs))
    basefont = TagProp('basefont', set(componentattrs + i18nattrs + ('id', 'size', 'color', 'face')))
    bdo = TagProp('bdo', set(componentattrs + eventattrs + ('lang', 'dir')))
    big = TagProp('big', set(allattrs))
    blockquote = TagProp('blockquote', set(allattrs + ('cite',)))
    body = TagProp('body', set(allattrs + (
        'onload', 'onunload', 'onfocus', 'background', 'bgcolor', 'text',
        'link', 'vlink', 'alink', 'leftmargin', 'topmargin', 'marginwidth', 'marginheight'
    )))
    br = TagProp('br', set(componentattrs + ('clear',)))
    button = TagProp('button', set(allattrs + focusattrs + ('name', 'value', 'type', 'disabled')))
    caption = TagProp('caption', set(allattrs + ('align',)))
    center = TagProp('center', set(allattrs))
    cite = TagProp('cite', set(allattrs))
    code = TagProp('code', set(allattrs))
    col = TagProp('col', set(allattrs + cellhalignattrs + cellvalignattrs + ('span', 'width')))
    colgroup = TagProp('colgroup', set(allattrs + cellhalignattrs + cellvalignattrs + ('span', 'width')))
    dd = TagProp('dd', set(allattrs))
    del_ = TagProp('del', set(allattrs + ('cite', 'datetime')))
    dfn = TagProp('dfn', set(allattrs))
    dir = TagProp('dir', set(allattrs + ('compact',)))
    div = TagProp('div', set(allattrs + ('align',)))
    dl = TagProp('dl', set(allattrs + ('compact',)))
    dt = TagProp('dt', set(allattrs))
    em = TagProp('em', set(allattrs))
    embed = TagProp('embed', set((
        'width', 'height', 'src', 'controller', 'href', 'target',
        'border', 'pluginspage', 'quality', 'type', 'bgcolor', 'menu'
    )))
    fieldset = TagProp('fieldset', set(allattrs))
    font = TagProp('font', set(componentattrs + i18nattrs + ('face', 'size', 'color')))
    form = TagProp('form', set(allattrs + (
        'action', 'method', 'name', 'enctype',
        'onsubmit', 'onreset', 'accept_charset', 'target'
    )))
    frame = TagProp('frame', set())
    frameset = TagProp('frameset', set(componentattrs + (
        'rows', 'cols', 'onload', 'onunload', 'framespacing', 'border',
        'marginwidth', 'marginheight', 'frameborder', 'noresize', 'scrolling'
    )))
    h1 = TagProp('h1', set(allattrs + ('align',)))
    h2 = TagProp('h2', set(allattrs + ('align',)))
    h3 = TagProp('h3', set(allattrs + ('align',)))
    h4 = TagProp('h4', set(allattrs + ('align',)))
    h5 = TagProp('h5', set(allattrs + ('align',)))
    h6 = TagProp('h6', set(allattrs + ('align',)))
    hr = TagProp('hr', set(allattrs + ('align', 'noshade', 'size', 'width', 'color')))
    html = TagProp('html', set(i18nattrs + ('id',)))
    i = TagProp('i', set(allattrs))
    iframe = TagProp('iframe', set(componentattrs + (
        'longdesc', 'name', 'src', 'frameborder', 'marginwidth', 'marginheight',
        'noresize', 'scrolling', 'align', 'height', 'width', 'hspace', 'vspace', 'bordercolor',
    )))
    img = TagProp('img', set(allattrs + (
        'src', 'alt', 'name', 'longdesc', 'width', 'height', 'usemap',
        'ismap', 'align', 'border', 'hspace', 'vspace', 'lowsrc'
    )))
    input = TagProp('input', set(allattrs + focusattrs + (
        'type', 'name', 'value', 'checked', 'disabled', 'readonly', 'size', 'maxlength',
        'src', 'alt', 'usemap', 'onselect', 'onchange', 'accept', 'align', 'border'
    )))
    ins = TagProp('ins', set(allattrs + ('cite', 'datetime')))
    isindex = TagProp('isindex', set(componentattrs + i18nattrs + ('prompt',)))
    kbd = TagProp('kbd', set(allattrs))
    label = TagProp('label', set(allattrs + ('for', 'accesskey', 'onfocus', 'onblur')))
    legend = TagProp('legend', set(allattrs + ('accesskey', 'align')))
    li = TagProp('li', set(allattrs + ('type', 'value')))
    map = TagProp('map', set(i18nattrs + eventattrs + ('id', 'class', 'style', 'title', 'name')))
    menu = TagProp('menu', set(allattrs + ('compact',)))
    noframes = TagProp('noframes', set(allattrs))
    noscript = TagProp('noscript', set(allattrs))
    object = TagProp('object', set(allattrs + (
        'declare', 'classid', 'codebase', 'data', 'type', 'codetype',
        'archive', 'standby', 'height', 'width', 'usemap', 'name',
        'tabindex', 'align', 'border', 'hspace', 'vspace'
    )))
    ol = TagProp('ol', set(allattrs + ('type', 'compact', 'start')))
    optgroup = TagProp('optgroup', set(allattrs + ('disabled', 'label')))
    option = TagProp('option', set(allattrs + ('selected', 'disabled', 'label', 'value')))
    p = TagProp('p', set(allattrs + ('align',)))
    param = TagProp('param', set(('id', 'name', 'value', 'valuetype', 'type')))
    pre = TagProp('pre', set(allattrs + ('width',)))
    q = TagProp('q', set(allattrs + ('cite',)))
    s = TagProp('s', set(allattrs))
    samp = TagProp('samp', set(allattrs))
    script = TagProp('script', set(('id', 'charset', 'type', 'language', 'src', 'defer')))
    select = TagProp('select', set(allattrs + (
        'name', 'size', 'multiple', 'disabled', 'tabindex',
        'onfocus', 'onblur', 'onchange', 'rows'
    )))
    small = TagProp('small', set(allattrs))
    span = TagProp('span', set(allattrs))
    strike = TagProp('strike', set(allattrs))
    strong = TagProp('strong', set(allattrs))
    style = TagProp('style', set(i18nattrs + ('id', 'type', 'media', 'title')))
    sub = TagProp('sub', set(allattrs))
    sup = TagProp('sup', set(allattrs))
    table = TagProp('table', set(componentattrs + i18nattrs + ('prompt',)))
    tbody = TagProp('tbody', set(allattrs + cellhalignattrs + cellvalignattrs))
    td = TagProp('td', set(allattrs + cellhalignattrs + cellvalignattrs + (
        'abbr', 'axis', 'headers', 'scope', 'rowspan',
        'colspan', 'nowrap', 'bgcolor', 'width', 'height',
        'background', 'bordercolor'
    )))
    textarea = TagProp('textarea', set(allattrs + focusattrs + (
        'name', 'rows', 'cols', 'disabled',
        'readonly', 'onselect', 'onchange', 'wrap'
    )))
    tfoot = TagProp('tfoot', set(allattrs + cellhalignattrs + cellvalignattrs))
    th = TagProp('th', set(allattrs + cellhalignattrs + cellvalignattrs + (
        'abbr', 'axis', 'headers', 'scope', 'rowspan',
        'colspan', 'nowrap', 'bgcolor', 'width', 'height'
        'background', 'bordercolor'
    )))
    thead = TagProp('thead', set(allattrs + cellhalignattrs + cellvalignattrs))
    tr = TagProp('tr', set(allattrs + cellhalignattrs + cellvalignattrs + ('bgcolor', 'nowrap', 'width', 'background')))
    tt = TagProp('tt', set(allattrs))
    u = TagProp('u', set(allattrs))
    ul = TagProp('ul', set(allattrs + ('type', 'compact')))
    var = TagProp('var', set(allattrs))

#    frame = TagProp('frame', set(componentattrs + ('longdesc', 'name', 'src', 'frameborder', 'marginwidht', 'marginheight',
#                                                                  'noresize', 'scrolling', 'framespacing', 'border', 'marginwidth', 'marginheight',
#                                                                  'frameborder', 'noresize', 'scrolling')))
    @classmethod
    def class_init(cls, specialTags):
        """Class initialisation

        In:
          -- ``special_tags`` -- tags that have a special factory
        """
        # Create a HTML parser that generate ``_HTMLTag`` nodes
        cls._html_parser = ET.HTMLParser()
        cls._html_parser.setElementClassLookup(ET.ElementDefaultClassLookup(element=_HTMLTag))

    def __init__(self, parent=None, **kw):
        """Renderer initialisation

        In:
          - ``parent`` -- parent renderer
        """
        super(Renderer, self).__init__(parent)

        if parent is None:
            self.head = self.head_renderer_factory(**kw)
        else:
            self.head = parent.head

    def makeelement(self, tag):
        """Make a tag

        In:
          - ``tag`` -- name of the tag to create

        Return:
          - the new tag
        """
        return self._makeelement(tag, self._html_parser)

    def _parse_html(self, parser, source, fragment, no_leading_text, **kw):
        """Parse a HTML file

        In:
          - ``parser`` -- XML or HTML parser to use
          - ``source`` -- can be a filename or a file object
          - ``fragment`` -- if ``True``, can parse a HTML fragment i.e a HTML without
            a unique root
          - ``no_leading_text`` -- if ``fragment`` is ``True``, ``no_leading_text``
            is ``False`` and the HTML to parsed begins by a text, this text is keeped
          - ``kw`` -- keywords parameters are passed to the HTML parser

        Return:
          - the root element of the parsed HTML, if ``fragment`` is ``False``
          - a list of HTML elements, if ``fragment`` is ``True``
        """
        if isinstance(source, basestring):
            if source.startswith(('http://', 'https://', 'ftp://')):
                source = urllib.urlopen(source)
            else:
                source = open(source)

        if not fragment:
            # Parse a HTML file
            # ----------------

            root = ET.parse(source, parser).getroot()
            source.close()

            # Attach the renderer to the root
            root._renderer = self
            return root

        # Parse a HTML fragment
        # ---------------------

        # Create a dummy ``<body>``
        html = cStringIO.StringIO('<html><body>%s</body></html>' % source.read())
        source.close()

        body = ET.parse(html, parser).getroot()[0]
        for e in body:
            if isinstance(e, _HTMLTag):
                # Attach the renderer to each roots
                e._renderer = self

        # Return the children of the dummy ``<body>``
        return ([body.text] if body.text and not no_leading_text else []) + body[:]

    def parse_html(self, source, fragment=False, no_leading_text=False, xhtml=False, **kw):
        """Parse a (X)HTML file

        In:
          - ``source`` -- can be a filename or a file object
          - ``fragment`` -- if ``True``, can parse a HTML fragment i.e a HTML without
            a unique root
          - ``no_leading_text`` -- if ``fragment`` is ``True``, ``no_leading_text``
            is ``False`` and the HTML to parsed begins by a text, this text is keeped
          - ``xhtml`` -- is the HTML to parse a valid XHTML ?
          - ``kw`` -- keywords parameters are passed to the HTML parser

        Return:
          - the root element of the parsed HTML, if ``fragment`` is ``False``
          - a list of HTML elements, if ``fragment`` is ``True``
        """
        parser = ET.XMLParser(**kw) if xhtml else ET.HTMLParser(**kw)
        parser.setElementClassLookup(ET.ElementDefaultClassLookup(element=_HTMLTag))

        return self._parse_html(parser, source, fragment, no_leading_text, **kw)

    def parse_htmlstring(self, text, fragment=False, no_leading_text=False, xhtml=False, **kw):
        """Parse a (X)HTML string

        In:
          - ``text`` -- can be a ``str`` or ``unicode`` strings
          - ``source`` -- can be a filename or a file object
          - ``fragment`` -- if ``True``, can parse a HTML fragment i.e a HTML without
            a unique root
          - ``no_leading_text`` -- if ``fragment`` is ``True``, ``no_leading_text``
            is ``False`` and the HTML to parsed begins by a text, this text is keeped
          - ``xhtml`` -- is the HTML to parse a valid XHTML ?
          - ``kw`` -- keywords parameters are passed to the HTML parser

        Return:
          - the root element of the parsed HTML, if ``fragment`` is ``False``
          - a list of HTML elements, if ``fragment`` is ``True``
        """
        if isinstance(text, unicode):
            text = text.encode(kw.setdefault('encoding', 'utf-8'))

        return self.parse_html(cStringIO.StringIO(text), fragment, no_leading_text, xhtml, **kw)


# ---------------------------------------------------------------------------

if __name__ == '__main__':
    t = ((1, 'a'), (2, 'b'), (3, 'c'))

    h = Renderer()

    h.head << h.head.title('A test')
    h.head << h.head.script('function() {}')
    h.head << h.head.meta(name='meta1', content='content1')

    with h.body(onload='javascript:alert()'):
        with h.ul:
            with h.li('Hello'):
                pass
            with h.li:
                h << 'world'
            h << h.li('yeah')

        with h.div(class_='toto'):
            with h.h1('moi'):
                h << h.i('toto')

        with h.div:
            h << 'yeah'
            for i in range(3):
                h << i

        with h.table(toto='toto'):
            for row in t:
                with h.tr:
                    for column in row:
                        with h.td:
                            h << column

    print h.html(h.head.head(h.head.root), h.root).write_htmlstring(pretty_print=True)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``create`` administrative command

Generate the directory structure and the skeleton of an application.
"""

from __future__ import with_statement

import sys
import os
import textwrap
import re
from StringIO import StringIO

from configobj import ConfigObj
from nagare.admin import command


def set_options(optparser):
    optparser.usage += ' <application>'


def create_empty_file(filename):
    open(filename, 'w').close()


def create_or_update(path, create=None, update=None, *args, **kw):
    if not create:
        def create(*args, **kw):
            pass

    if not update:
        def update(*args, **kw):
            pass

    return (update if os.path.exists(path) else create)(path, *args, **kw)


def create_setup_py(filename, params):
    upgrade_setup_py(
        filename,
        params,
        textwrap.dedent('''\
        from setuptools import setup, find_packages

        VERSION = '0.0.1'


        setup(
            name='%(name)s',
            version=VERSION,
            author='',
            author_email='',
            description='',
            license='',
            keywords='',
            url='',
            packages=find_packages(),
            include_package_data=True,
            package_data={'': ['*.cfg']},
            zip_safe=False,
            install_requires=('nagare',),
            entry_points="""
            [nagare.applications]
            %(id)s = %(id)s.app:app
            """
        )''' % params)
    )


def upgrade_setup_py(filename, params, setup_py=None):
    if not setup_py:
        with open(filename, 'r') as f:
            setup_py = f.read()
            if 'message_extractors' in setup_py:
                return

    if os.path.exists(filename):
        os.rename(filename, filename + '.old')

    with open(filename, 'w') as f:
        print >>f, re.sub(r'\n(\s*)(entry_points)', r"\n\1message_extractors={'%(id)s': [('**.py', 'python', None)]},\n\1\2" % params, setup_py)


def upgrade_setup_cfg(filename):
    old_conf = ConfigObj(filename)

    new_conf = ConfigObj(StringIO(textwrap.dedent('''\
            [extract_messages]
            keywords = _ , _N:1,2 , _L , _LN:1,2 , gettext , ugettext , ngettext:1,2 , ungettext:1,2 , lazy_gettext , lazy_ugettext , lazy_ngettext:1,2 , lazy_ungettext:1,2
            output_file = data/locale/messages.pot

            [init_catalog]
            input_file = data/locale/messages.pot
            output_dir = data/locale
            domain = messages

            [update_catalog]
            input_file = data/locale/messages.pot
            output_dir = data/locale
            domain = messages

            [compile_catalog]
            directory = data/locale
            domain = messages
            ''')), list_values=False)
    new_conf.merge(old_conf)

    if new_conf != old_conf:
        if os.path.exists(filename):
            os.rename(filename, filename + '.old')

        new_conf.filename = filename
        new_conf.write()


def create_manifest_in(filename):
    with open(filename, 'w') as f:
        print >>f, textwrap.dedent('''\
            graft conf
            graft static
            graft data/locale
        ''')


def create_app_py(filename, params):
    with open(filename, 'w') as f:
        f.write(textwrap.dedent("""\
            from __future__ import with_statement

            import os
            from nagare import presentation


            class %(Id)s(object):
                pass


            @presentation.render_for(%(Id)s)
            def render(self, h, *args):
                NAGARE = 'http://www.nagare.org/'
                NAGARE_DOC = NAGARE + 'doc/'

                this_file = __file__
                if this_file.endswith('.pyc'):
                    this_file = __file__[:-1]

                models_file = os.path.join(os.path.dirname(__file__), 'models.py')

                h.head << h.head.title('Up and Running !')

                h.head.css_url('/static/nagare/application.css')
                h.head.css(
                    'defaultapp',
                    '''#main {
                           margin-left: 20px;
                           padding-bottom: 100px;
                           background: url(/static/nagare/img/sakura.jpg) no-repeat 123px 100%%;
                    }'''  # noqa: E501
                )

                with h.div(id='body'):
                    h << h.a(
                        h.img(src='/static/nagare/img/logo.png'),
                        id='logo',
                        href=NAGARE, title='Nagare home'
                    )

                    with h.div(id='content'):
                        h << h.div('Congratulations!', id='title')

                        with h.div(id='main'):
                            h << h.h1('Your application is running')

                            h << 'You can now:'
                            with h.ul:
                                with h.li:
                                    h << 'If your application uses a database, '
                                    h << 'add your database entities into '
                                    h << h.em(models_file)
                                with h.li:
                                    h << 'Add your application components into ',
                                    h << h.em(this_file)
                                    h << ' or create new files'

                            h << h.em("Have fun!")

                with h.div(id='footer'):
                    with h.table:
                        with h.tr:
                            h << h.th('About Nagare')
                            h << h.th('Community')
                            h << h.th('Learn', class_='last')

                        with h.tr:
                            with h.td:
                                with h.ul:
                                    h << h.li(h.a('Description', href=NAGARE_DOC+'description'))  # noqa: E501
                                    h << h.li(h.a('Features', href=NAGARE_DOC+'features'))  # noqa: E501
                                    h << h.li(h.a('License', href=NAGARE_DOC+'license'))  # noqa: E501

                            with h.td:
                                with h.ul:
                                    h << h.li(h.a('Github repositories', href='http://github.com/nagareproject'))  # noqa: E501
                                    h << h.li(h.a('Mailing list', href='http://groups.google.com/group/nagare-users'))  # noqa: E501
                                    h << h.li(h.a('Bugs report', href='https://github.com/nagareproject/core/issues'))  # noqa: E501

                            with h.td(class_='last'):
                                with h.ul:
                                    h << h.li(h.a('Documentation', href=NAGARE_DOC))  # noqa: E501
                                    h << h.li(h.a('Demonstrations portal', href=NAGARE+'portal'))  # noqa: E501
                                    h << h.li(h.a('Demonstrations', href=NAGARE+'demo'))  # noqa: E501
                                    h << h.li(h.a('Wiki Tutorial', href=NAGARE+'wiki'))  # noqa: E501

                return h.root


            # ---------------------------------------------------------------

            app = %(Id)s
        """ % params))


def create_models_py(filename):
    with open(filename, 'w') as f:
        f.write(textwrap.dedent('''\
            from elixir import *  # noqa: F403, F401
            from sqlalchemy import MetaData

            __metadata__ = MetaData()

            # Here, put the definition of your Elixir or SQLAlchemy models
        '''))


def create_conf(filename, params):
    with open(filename, 'w') as f:
        f.write(textwrap.dedent('''\
            [application]
            path = app %(id)s
            name = %(id)s
            debug = off

            [database]
            activated = off
            uri = sqlite:///$here/../data/%(id)s.db
            metadata = %(id)s.models:__metadata__
            debug = off
        ''' % params))


def run(parser, options, args):
    if len(args) == 0:
        parser.error('an application name must be provided')

    if len(args) > 1:
        parser.error('only one application name must be provided')

    root = args[0]
    app_name = os.path.basename(root)
    app_id = app_name.replace('-', '_')

    create_or_update(root, os.mkdir)

    params = {
        'exe': sys.executable,
        'root': root,
        'setup': os.path.join(root, 'setup.py'),
        'name': app_name,
        'id': app_id,
        'Id': app_id.capitalize()
    }

    create_or_update(os.path.join(root, 'setup.py'), create_setup_py, upgrade_setup_py, params)
    create_or_update(os.path.join(root, 'setup.cfg'), upgrade_setup_cfg, upgrade_setup_cfg)
    create_or_update(os.path.join(root, 'MANIFEST.in'), create_manifest_in)

    create_or_update(os.path.join(root, app_id), os.mkdir)
    create_or_update(os.path.join(root, app_id, '__init__.py'), create_empty_file)
    create_or_update(os.path.join(root, app_id, 'app.py'), create_app_py, None, params)
    create_or_update(os.path.join(root, app_id, 'models.py'), create_models_py)

    create_or_update(os.path.join(root, 'static'), os.mkdir)
    create_or_update(os.path.join(root, 'static', '__init__.py'), create_empty_file)

    create_or_update(os.path.join(root, 'data'), os.mkdir)
    create_or_update(os.path.join(root, 'data', '__init__.py'), create_empty_file)
    create_or_update(os.path.join(root, 'data', 'locale'), os.mkdir)

    create_or_update(os.path.join(root, 'conf'), os.mkdir)
    create_or_update(os.path.join(root, 'conf', '__init__.py'), create_empty_file)
    create_or_update(os.path.join(root, 'conf', app_id + '.cfg'), create_conf, None, params)

    print "Application '%s' created." % app_name
    print
    print textwrap.dedent("""\
        Note:
          1. Edit the file '%(setup)s' to set the informations about your new application.
          2. Register your application with:
               - cd "%(root)s"
               - "%(exe)s" setup.py develop
        """ % params)


class Create(command.Command):
    desc = 'Create an application skeleton'

    set_options = staticmethod(set_options)
    run = staticmethod(run)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Base class of all the publishers"""

import random

from paste import fileapp, httpexceptions, urlmap
import configobj

from nagare import config


def serve_file(filename):
    """Create a WSGI application that return a static file

   In:
     - ``filename`` -- path of the file to serve

   Return:
     - a WSGI application
   """
    return httpexceptions.HTTPNotFound() if filename is None else fileapp.FileApp(filename)


class Publisher(object):
    """Base class of all the publishers"""

    default_port = 8080  # Port to listen to
    spec = {}  # Command line options

    def __init__(self):
        """Initialisation
       """
        self.urls = urlmap.URLMap()  # Dispatch for all the registered applications
        self.apps = {}  # List of all the registered applications

    def register_application(self, app_name, name, app, wsgi_pipe):
        """Register an application to launch

       In:
         - ``app_name`` -- name of the application
         - ``name`` -- url of the application
         - ``app`` -- the WSGI application
         - ``wsgi_pipe`` -- the pipe of WSGI "middlewares"
       """
        self.apps.setdefault(app, (app_name, []))
        self.apps[app][1].append(name)

        self.urls['/' + name] = wsgi_pipe
        print "Application '%s' registered as '/%s'" % (app_name, name)

    def get_registered_applications(self):
        """Return all the applications registered on this publisher
       """
        return [(app, app_path, app_urls) for (app, (app_path, app_urls)) in self.apps.items()]

    def register_static(self, name, get_file):
        """Register a WSGI application to serve static contents

       In:
         - ``name`` -- the URL of the contents will be prefix by ``/static/<name>/``
         - ``get_file`` -- function that will received the URL of the static content
           and will return its filename

       Return:
         - URL prefix (``/static/<name>/``)
       """
        self.urls['/static/' + name] = lambda environ, start_response: serve_file(get_file(environ['PATH_INFO']))(environ, start_response)

        return '/static/' + name + '/'

    def _validate_conf(self, filename, conf, error):
        """Validate the configuration read from the publisher configuration file

       In:
         -  ``filename`` -- the path to the configuration file
         - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
         - ``error`` -- the function to call in case of configuration errors

       Return:
         - the tuple:

           - hostname to listen to
           - port to listen to
           - conf object
       """
        conf = dict([(k, v) for (k, v) in conf.items() if k in self.spec])
        conf = configobj.ConfigObj(conf, configspec=self.spec, interpolation='Template')
        config.validate(filename, conf, error)
        conf = dict([(k, v) for (k, v) in conf.items() if v is not None])

        return conf.pop('host'), conf.pop('port'), conf

    def on_new_process(self):
        """The publisher has started a new process
       """
        random.seed(None)

        # Notify each registered applications
        for app in self.apps:
            app.start()

    def serve(self, filename, conf, error):
        """Run the publisher

       In:
         -  ``filename`` -- the path to the configuration file
         - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
         - ``error`` -- the function to call in case of configuration errors
       """
        pass



# Encoding: utf-8

# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

import decimal
import datetime
import unittest

import pytz

from nagare import local, i18n


class Translation(dict):
    def gettext(self, msg):
        return self[msg]

    def ugettext(self, msg):
        return unicode(self.gettext(msg))

    def ngettext(self, singular, plural, n):
        return self[singular if n == 1 else plural]

    def ungettext(self, singular, plural, n):
        return unicode(self.ngettext(singular, plural, n))


class Locale(i18n.Locale):
    def _get_translation(self, domain=None):
        return Translation({
            'hello': 'bonjour',
            'Holidays': 'Vacances %(year)d',
            'horse': 'cheval',
            'horses': 'chevaux'
        })


# -----------------------------------------------------------------------------

class TestTranslations(unittest.TestCase):
    def setUp(self):
        local.request = local.Process()
        i18n.set_locale(Locale('fr', 'FR'))

    def test_gettext(self):
        s = i18n.gettext('hello')
        self.assertTrue(isinstance(s, str))
        self.assertEqual(s, 'bonjour')

    def test_gettext_params(self):
        s = i18n.gettext('Holidays', year=2010)
        self.assertIsInstance(s, str)
        self.assertEqual(s, 'Vacances 2010')

    def test_gettext_unknown(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))
        s = i18n.gettext('unknown')
        self.assertIsInstance(s, str)
        self.assertEqual(s, 'unknown')

    def test_ugettext(self):
        s = i18n.ugettext('hello')
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'bonjour')

    def test_ugettext_params(self):
        s = i18n.ugettext('Holidays', year=2010)
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'Vacances 2010')

    def test_ugettext_unknown(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))
        s = i18n.ugettext('unknown')
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'unknown')

    def test__(self):
        s = i18n._('hello')
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'bonjour')

    def test_ngettext_singular(self):
        s = i18n.ngettext('horse', 'horses', 1)
        self.assertIsInstance(s, str)
        self.assertEqual(s, 'cheval')

    def test_ngettext_plural(self):
        s = i18n.ngettext('horse', 'horses', 3)
        self.assertIsInstance(s, str)
        self.assertEqual(s, 'chevaux')

    def test_ngettext_singular_unknown(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))
        s = i18n.ngettext('unknown1', 'unknown2', 1)
        self.assertIsInstance(s, str)
        self.assertEqual(s, 'unknown1')

    def test_ngettext_plural_unknown(self):
        i18n.set_locale(i18n.Locale('fr_FR'))
        s = i18n.ngettext('unknown1', 'unknown2', 3)
        self.assertIsInstance(s, str)
        self.assertEqual(s, 'unknown2')

    def test_ungettext_singular(self):
        s = i18n.ungettext('horse', 'horses', 1)
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'cheval')

    def test_ungettext_plural(self):
        s = i18n.ungettext('horse', 'horses', 3)
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'chevaux')

    def test_ungettext_singular_unknown(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))
        s = i18n.ungettext('unknown1', 'unknown2', 1)
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'unknown1')

    def test_ungettext_plural_unknown(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))
        s = i18n.ungettext('unknown1', 'unknown2', 3)
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'unknown2')

    def test_N_singular(self):
        s = i18n._N('horse', 'horses', 1)
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'cheval')

    def test_N_plural(self):
        s = i18n._N('horse', 'horses', 3)
        self.assertIsInstance(s, unicode)
        self.assertEqual(s, u'chevaux')

    def test_lazy_gettext(self):
        s = i18n.lazy_gettext('hello')
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, str)
        self.assertEqual(s, 'bonjour')

    def test_lazy_gettext_params(self):
        s = i18n.lazy_gettext('Holidays', year=2010)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, str)
        self.assertEqual(s, 'Vacances 2010')

    def test_lazy_ugettext(self):
        s = i18n.lazy_ugettext('hello')
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, unicode)
        self.assertEqual(s, u'bonjour')

    def test_lazy_ugettext_params(self):
        s = i18n.lazy_ugettext('Holidays', year=2010)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, unicode)
        self.assertEqual(s, u'Vacances 2010')

    def test_L_ugettext(self):
        s = i18n._L('hello')
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, unicode)
        self.assertEqual(s, u'bonjour')

    def test_lazy_ngettext_singular(self):
        s = i18n.lazy_ngettext('horse', 'horses', 1)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, str)
        self.assertEqual(s, 'cheval')

    def test_lazy_ngettext_plural(self):
        s = i18n.lazy_ngettext('horse', 'horses', 3)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, str)
        self.assertEqual(s, 'chevaux')

    def test_lazy_ungettext_singular(self):
        s = i18n.lazy_ungettext('horse', 'horses', 1)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, unicode)
        self.assertEqual(s, u'cheval')

    def test_lazy_ungettext_plural(self):
        s = i18n.lazy_ungettext('horse', 'horses', 3)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, unicode)
        self.assertEqual(s, u'chevaux')

    def test_LN_ungettext_singular(self):
        s = i18n._LN('horse', 'horses', 1)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, unicode)
        self.assertEqual(s, u'cheval')

    def test_LN_ungettext_plural(self):
        s = i18n._LN('horse', 'horses', 3)
        self.assertEqual(s.__class__.__name__, 'LazyProxy')
        self.assertIsInstance(s.value, unicode)
        self.assertEqual(s, u'chevaux')


# -----------------------------------------------------------------------------

class TestDateNames(unittest.TestCase):
    def setUp(self):
        local.request = local.Process()
        i18n.set_locale(i18n.Locale('fr', 'FR'))

    def test_get_period_names(self):
        self.assertEqual(i18n.get_period_names(), {
            'afternoon1': u'après-midi',
            'am': u'AM',
            'evening1': u'soir',
            'midnight': u'minuit',
            'morning1': u'matin',
            'night1': u'nuit',
            'noon': u'midi',
            'pm': u'PM'
        })

    def test_get_day_names(self):
        self.assertEqual(i18n.get_day_names(), {0: u'lundi', 1: u'mardi', 2: u'mercredi', 3: u'jeudi', 4: u'vendredi', 5: u'samedi', 6: u'dimanche'})
        self.assertEqual(i18n.get_day_names(width='wide'), {0: u'lundi', 1: u'mardi', 2: u'mercredi', 3: u'jeudi', 4: u'vendredi', 5: u'samedi', 6: u'dimanche'})
        self.assertEqual(i18n.get_day_names(width='abbreviated'), {0: u'lun.', 1: u'mar.', 2: u'mer.', 3: u'jeu.', 4: u'ven.', 5: u'sam.', 6: u'dim.'})
        self.assertEqual(i18n.get_day_names(width='narrow'), {0: u'L', 1: u'M', 2: u'M', 3: u'J', 4: u'V', 5: u'S', 6: u'D'})

    def test_get_month_names(self):
        self.assertEqual(i18n.get_month_names(), {1: u'janvier', 2: u'février', 3: u'mars', 4: u'avril', 5: u'mai', 6: u'juin', 7: u'juillet', 8: u'août', 9: u'septembre', 10: u'octobre', 11: u'novembre', 12: u'décembre'})
        self.assertEqual(i18n.get_month_names(width='wide'), {1: u'janvier', 2: u'février', 3: u'mars', 4: u'avril', 5: u'mai', 6: u'juin', 7: u'juillet', 8: u'août', 9: u'septembre', 10: u'octobre', 11: u'novembre', 12: u'décembre'})
        self.assertEqual(i18n.get_month_names(width='abbreviated'), {1: u'janv.', 2: u'févr.', 3: u'mars', 4: u'avr.', 5: u'mai', 6: u'juin', 7: u'juil.', 8: u'août', 9: u'sept.', 10: u'oct.', 11: u'nov.', 12: u'déc.'})
        self.assertEqual(i18n.get_month_names(width='narrow'), {1: u'J', 2: u'F', 3: u'M', 4: u'A', 5: u'M', 6: u'J', 7: u'J', 8: u'A', 9: u'S', 10: u'O', 11: u'N', 12: u'D'})

    def test_get_quarter_names(self):
        self.assertEqual(i18n.get_quarter_names(), {1: u'1er trimestre', 2: u'2e trimestre', 3: u'3e trimestre', 4: u'4e trimestre'})
        self.assertEqual(i18n.get_quarter_names(width='wide'), {1: u'1er trimestre', 2: u'2e trimestre', 3: u'3e trimestre', 4: u'4e trimestre'})
        self.assertEqual(i18n.get_quarter_names(width='abbreviated'), {1: u'T1', 2: u'T2', 3: u'T3', 4: u'T4'})
        self.assertEqual(i18n.get_quarter_names(width='narrow'), {1: u'1', 2: u'2', 3: u'3', 4: u'4'})

    def test_get_era_names(self):
        i18n.set_locale(i18n.Locale('en', 'US'))

        self.assertEqual(i18n.get_era_names(), {0: u'Before Christ', 1: u'Anno Domini'})
        self.assertEqual(i18n.get_era_names(width='wide'), {0: u'Before Christ', 1: u'Anno Domini'})
        self.assertEqual(i18n.get_era_names(width='abbreviated'), {0: u'BC', 1: u'AD'})
        self.assertEqual(i18n.get_era_names(width='narrow'), {0: u'B', 1: u'A'})

    def test_get_date_format(self):
        self.assertEqual(i18n.get_date_format(format='full').pattern, 'EEEE d MMMM y')
        self.assertEqual(i18n.get_date_format(format='long').pattern, 'd MMMM y')
        self.assertEqual(i18n.get_date_format().pattern, 'd MMM y')
        self.assertEqual(i18n.get_date_format(format='medium').pattern, 'd MMM y')
        self.assertEqual(i18n.get_date_format(format='short').pattern, 'dd/MM/y')

    def test_get_datetime_format(self):
        self.assertEqual(i18n.get_datetime_format(format='full'), u"{1} 'à' {0}")
        self.assertEqual(i18n.get_datetime_format(format='long'), u"{1} 'à' {0}")
        self.assertEqual(i18n.get_datetime_format(), u"{1} 'à' {0}")
        self.assertEqual(i18n.get_datetime_format(format='medium'), u"{1} 'à' {0}")
        self.assertEqual(i18n.get_datetime_format(format='short'), u'{1} {0}')

    def test_get_time_format(self):
        self.assertEqual(i18n.get_time_format(format='full').pattern, 'HH:mm:ss zzzz')
        self.assertEqual(i18n.get_time_format(format='long').pattern, 'HH:mm:ss z')
        self.assertEqual(i18n.get_time_format().pattern, 'HH:mm:ss')
        self.assertEqual(i18n.get_time_format(format='medium').pattern, 'HH:mm:ss')
        self.assertEqual(i18n.get_time_format(format='short').pattern, 'HH:mm')

    def test_get_timezone_gmt(self):
        utc_date = datetime.datetime(2007, 4, 1, 15, 30)

        tz = pytz.timezone('Pacific/Pitcairn')
        d = Locale('en', timezone=tz).to_timezone(utc_date)

        self.assertEqual(i18n.get_timezone_gmt(d, width='long'), u'UTC-08:00')
        self.assertEqual(i18n.get_timezone_gmt(d), u'UTC-08:00')
        self.assertEqual(i18n.get_timezone_gmt(d, width='short'), '-0800')

        i18n.set_locale(i18n.Locale('en', 'US'))

        self.assertEqual(i18n.get_timezone_gmt(d, width='long'), u'GMT-08:00')
        self.assertEqual(i18n.get_timezone_gmt(d), u'GMT-08:00')
        self.assertEqual(i18n.get_timezone_gmt(d, width='short'), '-0800')

    def test_get_timezone_location(self):
        tz = pytz.timezone('Africa/Bamako')
        self.assertEqual(i18n.get_timezone_location(tz), 'Mali')

    def test_get_timezone_name(self):
        utc_date = datetime.datetime(2007, 4, 1, 15, 30)

        tz = pytz.timezone('Pacific/Pitcairn')
        d = Locale('en', timezone=tz).to_timezone(utc_date)

        self.assertEqual(i18n.get_timezone_name(d), u'heure des îles Pitcairn')
        self.assertEqual(i18n.get_timezone_name(d, width='long'), u'heure des îles Pitcairn')
        self.assertEqual(i18n.get_timezone_name(d, width='short'), '-0800')


# -----------------------------------------------------------------------------

class TestFormats(unittest.TestCase):
    def setUp(self):
        local.request = local.Process()
        i18n.set_locale(i18n.Locale('fr', 'FR'))

    def test_get_currency_name(self):
        self.assertEqual(i18n.get_currency_name('USD'), u'dollar des États-Unis')

    def test_get_currency_symbol(self):
        self.assertEqual(i18n.get_currency_symbol('USD'), '$US')

    def test_get_decimal_symbol(self):
        self.assertEqual(i18n.get_decimal_symbol(), ',')

    def test_get_plus_sign_symbol(self):
        self.assertEqual(i18n.get_plus_sign_symbol(), '+')

    def test_get_minus_sign_symbol(self):
        self.assertEqual(i18n.get_minus_sign_symbol(), '-')

    def test_get_exponential_symbol(self):
        self.assertEqual(i18n.get_exponential_symbol(), 'E')

    def test_get_group_symbol(self):
        self.assertEqual(i18n.get_group_symbol(), u'\N{NO-BREAK SPACE}')

    def test_format_number(self):
        self.assertEqual(i18n.format_number(1099), u'1\N{NO-BREAK SPACE}099')

    def test_format_decimal(self):
        self.assertEqual(i18n.format_decimal(1236.1236), u'1\N{NO-BREAK SPACE}236,124')

    def test_format_currency(self):
        self.assertEqual(i18n.format_currency(1236.126, 'EUR'), u'1\N{NO-BREAK SPACE}236,13\N{NO-BREAK SPACE}\N{EURO SIGN}')

    def test_format_percent(self):
        self.assertEqual(i18n.format_percent(24.1234), u'2\N{NO-BREAK SPACE}412\N{NO-BREAK SPACE}%')

    def test_format_scientific(self):
        self.assertEqual(i18n.format_scientific(10000), '1E4')
        self.assertEqual(i18n.format_scientific(1234567, u'##0E00'), '1.23E06')

    def test_parse_number(self):
        self.assertEqual(i18n.parse_number(u'1\N{NO-BREAK SPACE}099'), 1099)

    def test_parse_decimal(self):
        self.assertEqual(i18n.parse_decimal(u'1\N{NO-BREAK SPACE}099,1234'), decimal.Decimal('1099.1234'))


# -----------------------------------------------------------------------------

class TestTimezones(unittest.TestCase):
    def test_to_timezone_no_timezone_datetime(self):
        d1 = datetime.datetime(2007, 4, 1, 15, 30)

        i18n.set_locale(i18n.Locale('fr', 'FR'))
        d2 = i18n.to_timezone(d1)
        self.assertIsNone(d2.tzinfo)
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Pacific/Pitcairn'))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'Pacific/Pitcairn')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Pacific/Pitcairn', default_timezone=pytz.UTC))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'Pacific/Pitcairn')
        self.assertEqual(d2.strftime('%H:%M'), '07:30')

    def test_to_timezone_utc_datetime(self):
        d1 = datetime.datetime(2007, 4, 1, 15, 30, tzinfo=pytz.UTC)

        i18n.set_locale(i18n.Locale('fr', 'FR'))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Pacific/Pitcairn'))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'Pacific/Pitcairn')
        self.assertEqual(d2.strftime('%H:%M'), '07:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Pacific/Pitcairn', default_timezone=pytz.UTC))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'Pacific/Pitcairn')
        self.assertEqual(d2.strftime('%H:%M'), '07:30')

    def test_to_timezone_local_datetime(self):
        tz = pytz.timezone('Pacific/Pitcairn')
        d1 = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))

        i18n.set_locale(i18n.Locale('fr', 'FR'))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'Pacific/Pitcairn')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'Africa/Niamey')
        self.assertEqual(d2.strftime('%H:%M'), '00:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey', default_timezone=pytz.UTC))
        d2 = i18n.to_timezone(d1)
        self.assertEqual(str(d2.tzinfo), 'Africa/Niamey')
        self.assertEqual(d2.strftime('%H:%M'), '00:30')

    def test_to_utc_no_timezone_datetime(self):
        d1 = datetime.datetime(2007, 4, 1, 15, 30)

        i18n.set_locale(i18n.Locale('fr', 'FR'))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '14:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey', default_timezone=pytz.UTC))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

    def test_to_utc_utc_datetime(self):
        d1 = datetime.datetime(2007, 4, 1, 15, 30, tzinfo=pytz.UTC)

        i18n.set_locale(i18n.Locale('fr', 'FR'))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey', default_timezone=pytz.UTC))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '15:30')

    def test_to_utc_local_datetime(self):
        tz = pytz.timezone('Pacific/Pitcairn')
        d1 = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))

        i18n.set_locale(i18n.Locale('fr', 'FR'))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '23:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '23:30')

        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey', default_timezone=pytz.UTC))
        d2 = i18n.to_utc(d1)
        self.assertEqual(str(d2.tzinfo), 'UTC')
        self.assertEqual(d2.strftime('%H:%M'), '23:30')

    def test_format_time_time_fr1(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))

        t = datetime.time(15, 30)

        self.assertEqual(i18n.format_time(t, format='full'), '15:30:00 UTC+00:00')
        self.assertEqual(i18n.format_time(t, format='long'), '15:30:00 +0000')
        self.assertEqual(i18n.format_time(t, format='medium'), '15:30:00')
        self.assertEqual(i18n.format_time(t), '15:30:00')
        self.assertEqual(i18n.format_time(t, format='short'), '15:30')

    def test_format_time_time_fr2(self):
        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))

        t = datetime.time(15, 30)

        self.assertEqual(i18n.format_time(t, format='full'), u'15:30:00 heure normale d’Afrique de l’Ouest')
        self.assertEqual(i18n.format_time(t, format='long'), '15:30:00 +0100')
        self.assertEqual(i18n.format_time(t, format='medium'), '15:30:00')
        self.assertEqual(i18n.format_time(t), '15:30:00')
        self.assertEqual(i18n.format_time(t, format='short'), '15:30')

    def test_format_time_time_fr3(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))

        t = datetime.time(15, 30, tzinfo=pytz.timezone('Pacific/Pitcairn'))

        self.assertEqual(i18n.format_time(t, format='full'), '15:30:00 UTC+00:00')
        self.assertEqual(i18n.format_time(t, format='long'), '15:30:00 +0000')
        self.assertEqual(i18n.format_time(t, format='medium'), '15:30:00')
        self.assertEqual(i18n.format_time(t), '15:30:00')
        self.assertEqual(i18n.format_time(t, format='short'), '15:30')

    def test_format_time_time_fr4(self):
        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))

        t = datetime.time(15, 30, tzinfo=pytz.timezone('Pacific/Pitcairn'))

        self.assertEqual(i18n.format_time(t, format='full'), u'15:30:00 heure normale d’Afrique de l’Ouest')
        self.assertEqual(i18n.format_time(t, format='long'), '15:30:00 +0100')
        self.assertEqual(i18n.format_time(t, format='medium'), '15:30:00')
        self.assertEqual(i18n.format_time(t), '15:30:00')
        self.assertEqual(i18n.format_time(t, format='short'), '15:30')

    def test_format_time_time_en(self):
        i18n.set_locale(i18n.Locale('en', 'US', timezone='Pacific/Pitcairn'))

        t = datetime.time(15, 30)

        self.assertEqual(i18n.format_time(t, format='full'), '3:30:00 PM Pitcairn Time')
        self.assertEqual(i18n.format_time(t, format='long'), '3:30:00 PM -0800')
        self.assertEqual(i18n.format_time(t, format='medium'), '3:30:00 PM')
        self.assertEqual(i18n.format_time(t), '3:30:00 PM')
        self.assertEqual(i18n.format_time(t, format='short'), '3:30 PM')

    def test_format_time_time_with_format(self):
        i18n.set_locale(i18n.Locale('en', 'US', timezone='Pacific/Pitcairn'))

        t = datetime.time(15, 30)
        self.assertEqual(i18n.format_time(t, format="hh 'o''clock' a, zzzz"), "03 o'clock PM, Pitcairn Time")

        t = datetime.time(15, 30, tzinfo=pytz.timezone('Africa/Niamey'))
        self.assertEqual(i18n.format_time(t, format="hh 'o''clock' a, zzzz"), "03 o'clock PM, Pitcairn Time")

    def test_format_time_datetime_fr1(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))

        d = datetime.datetime(2007, 4, 1, 15, 30)

        self.assertEqual(i18n.format_time(d, format='full'), '15:30:00 UTC+00:00')
        self.assertEqual(i18n.format_time(d, format='long'), '15:30:00 +0000')
        self.assertEqual(i18n.format_time(d, format='medium'), '15:30:00')
        self.assertEqual(i18n.format_time(d), '15:30:00')
        self.assertEqual(i18n.format_time(d, format='short'), '15:30')

    def test_format_time_datetime_fr2(self):
        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))

        d = datetime.datetime(2007, 4, 1, 15, 30)

        self.assertEqual(i18n.format_time(d, format='full'), u'15:30:00 heure normale d’Afrique de l’Ouest')
        self.assertEqual(i18n.format_time(d, format='long'), '15:30:00 +0100')
        self.assertEqual(i18n.format_time(d, format='medium'), '15:30:00')
        self.assertEqual(i18n.format_time(d), '15:30:00')
        self.assertEqual(i18n.format_time(d, format='short'), '15:30')

    def test_format_time_datetime_fr3(self):
        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey', default_timezone=pytz.UTC))

        d = datetime.datetime(2007, 4, 1, 15, 30)

        self.assertEqual(i18n.format_time(d, format='full'), u'16:30:00 heure normale d’Afrique de l’Ouest')
        self.assertEqual(i18n.format_time(d, format='long'), '16:30:00 +0100')
        self.assertEqual(i18n.format_time(d, format='medium'), '16:30:00')
        self.assertEqual(i18n.format_time(d), '16:30:00')
        self.assertEqual(i18n.format_time(d, format='short'), '16:30')

    def test_format_time_datetime_fr4(self):
        i18n.set_locale(i18n.Locale('fr', 'FR'))

        tz = pytz.timezone('Pacific/Pitcairn')
        d = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))

        self.assertEqual(i18n.format_time(d, format='full'), '23:30:00 UTC+00:00')
        self.assertEqual(i18n.format_time(d, format='long'), '23:30:00 +0000')
        self.assertEqual(i18n.format_time(d, format='medium'), '23:30:00')
        self.assertEqual(i18n.format_time(d), '23:30:00')
        self.assertEqual(i18n.format_time(d, format='short'), '23:30')

    def test_format_time_datetime_fr5(self):
        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))

        tz = pytz.timezone('Pacific/Pitcairn')
        d = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))

        self.assertEqual(i18n.format_time(d, format='full'), u'00:30:00 heure normale d’Afrique de l’Ouest')
        self.assertEqual(i18n.format_time(d, format='long'), '00:30:00 +0100')
        self.assertEqual(i18n.format_time(d, format='medium'), '00:30:00')
        self.assertEqual(i18n.format_time(d), '00:30:00')
        self.assertEqual(i18n.format_time(d, format='short'), '00:30')

    def test_format_time_datetime_with_format(self):
        i18n.set_locale(i18n.Locale('en', 'US', timezone='Pacific/Pitcairn'))

        d = datetime.datetime(2007, 4, 1, 15, 30)
        self.assertEqual(i18n.format_time(d, format="hh 'o''clock' a, zzzz"), u"03 o'clock PM, Pitcairn Time")

        tz = pytz.timezone('Africa/Niamey')
        d = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))
        self.assertEqual(i18n.format_time(d, format="hh 'o''clock' a, zzzz"), u"06 o'clock AM, Pitcairn Time")

    def test_format_date_date(self):
        d = datetime.date(2007, 4, 1)

        self.assertEqual(i18n.format_date(d, format='full'), 'dimanche 1 avril 2007')
        self.assertEqual(i18n.format_date(d, format='long'), '1 avril 2007')
        self.assertEqual(i18n.format_date(d, format='medium'), '1 avr. 2007')
        self.assertEqual(i18n.format_date(d), '1 avr. 2007')
        self.assertEqual(i18n.format_date(d, format='short'), '01/04/2007')

    def test_format_date_datetime(self):
        tz = pytz.timezone('Pacific/Pitcairn')
        d = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))

        self.assertEqual(i18n.format_date(d, format='full'), 'dimanche 1 avril 2007')
        self.assertEqual(i18n.format_date(d, format='long'), '1 avril 2007')
        self.assertEqual(i18n.format_date(d, format='medium'), '1 avr. 2007')
        self.assertEqual(i18n.format_date(d), '1 avr. 2007')
        self.assertEqual(i18n.format_date(d, format='short'), '01/04/2007')

    def test_format_date_date_with_format(self):
        d = datetime.date(2007, 4, 1)

        self.assertEqual(i18n.format_date(d, 'EEE, MMM d, yy'), 'dim., avr. 1, 07')

    def test_format_datetime(self):
        i18n.set_locale(i18n.Locale('fr', 'FR', timezone='Africa/Niamey'))

        tz = pytz.timezone('Pacific/Pitcairn')
        d = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))

        self.assertEqual(i18n.format_datetime(d, format='full'), u'lundi 2 avril 2007 à 00:30:00 heure normale d’Afrique de l’Ouest')
        self.assertEqual(i18n.format_datetime(d, format='long'), u'2 avril 2007 à 00:30:00 +0100')
        self.assertEqual(i18n.format_datetime(d, format='medium'), u'2 avr. 2007 à 00:30:00')
        self.assertEqual(i18n.format_datetime(d), u'2 avr. 2007 à 00:30:00')
        self.assertEqual(i18n.format_datetime(d, format='short'), '02/04/2007 00:30')

    def test_format_datetime_with_format(self):
        i18n.set_locale(i18n.Locale('en', 'US', timezone='Pacific/Pitcairn'))

        d = datetime.datetime(2007, 4, 1, 15, 30)
        self.assertEqual(i18n.format_datetime(d, format="yyyy.MM.dd G 'at' HH:mm:ss zzz"), '2007.04.01 AD at 15:30:00 -0800')

        tz = pytz.timezone('Africa/Niamey')
        d = tz.localize(datetime.datetime(2007, 4, 1, 15, 30))
        self.assertEqual(i18n.format_datetime(d, format="yyyy.MM.dd G 'at' HH:mm:ss zzz"), '2007.04.01 AD at 06:30:00 -0800')


# -----------------------------------------------------------------------------

class TestParsing(unittest.TestCase):
    def setUp(self):
        local.request = local.Process()
        i18n.set_locale(i18n.Locale('fr', 'FR'))

    def test_parse_time_fr(self):
        t = i18n.parse_time('15:30:10')
        self.assertIsInstance(t, datetime.time)
        self.assertEqual((t.hour, t.minute, t.second), (15, 30, 10))

    def test_parse_time_en(self):
        i18n.set_locale(i18n.Locale('en', 'US'))

        t = i18n.parse_time('15:30:10')
        self.assertIsInstance(t, datetime.time)
        self.assertEqual((t.hour, t.minute, t.second), (15, 30, 10))

    def test_parse_date_fr(self):
        d = i18n.parse_date('4/1/04')
        self.assertIsInstance(d, datetime.date)
        self.assertEqual((d.year, d.month, d.day), (2004, 1, 4))

        d = i18n.parse_date('4/1/2004')
        self.assertIsInstance(d, datetime.date)
        self.assertEqual((d.year, d.month, d.day), (2004, 1, 4))

    def test_parse_date_en(self):
        i18n.set_locale(i18n.Locale('en', 'US'))

        d = i18n.parse_date('4/1/04')
        self.assertIsInstance(d, datetime.date)
        self.assertEqual((d.year, d.month, d.day), (2004, 4, 1))

        d = i18n.parse_date('4/1/2004')
        self.assertIsInstance(d, datetime.date)
        self.assertEqual((d.year, d.month, d.day), (2004, 4, 1))


# -----------------------------------------------------------------------------

class TestContext(unittest.TestCase):
    def setUp(self):
        local.request = local.Process()

    def test_context_manager(self):
        locale1 = i18n.Locale('fr', 'FR', domain='domain1')
        locale2 = i18n.Locale('fr', 'FR', domain='domain2')

        i18n.set_locale(locale1)
        self.assertEqual(i18n.get_locale().domain, 'domain1')

        with locale2:
            self.assertEqual(i18n.get_locale().domain, 'domain2')
            with locale2:
                self.assertEqual(i18n.get_locale().domain, 'domain2')

        self.assertEqual(i18n.get_locale().domain, 'domain1')



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Add ``<html><head><body>`` around a tree"""

import types

from lxml import etree

from nagare import presentation


def search_element(element_name, l):
    """Search an element with ``element_name`` name as the first element in ``l``

    Skip the comments and processing instructions at the start of ``l``

    In:
      - ``element_name`` -- name of the element to search
      - ``l`` -- list of elements

    Return:
      - if found: (position of the next element, element found)
      - else: (0, None)
    """
    for i, element in enumerate(l):
        if isinstance(element, (etree._Comment, etree._ProcessingInstruction)):
            continue

        if isinstance(element, etree._Element) and element.tag.endswith(element_name):
            return i + 1, element

    return 0, None


def wrap(content_type, h, content):
    """Add the tags ``<html>``, ``<head>`` and ``<body>`` is they don't exist

    In:
      - ``content_type`` -- the content type to send to the browser
      - ``h`` -- the current renderer
      - ``content`` -- the rendered tree

    Return:
      - new tree with ``<html>``, ``<head>`` and ``<body>``
    """
    if 'html' not in content_type:
        return content

    if h.response.xml_output:
        h.namespaces = {None: 'http://www.w3.org/1999/xhtml'}

    if not isinstance(content, (list, tuple, types.GeneratorType)):
        content = [content]

    i, html = search_element('html', content)
    if html is None:
        html = content

    j, head = search_element('head', html)
    _, body = search_element('body', html[j:])

    if body is None:
        # No ``<body>`` found, add it
        html[j:] = [h.body(html[j:])]

    if head is None:
        # No ``<head>`` found, add it
        head = h.head.head
        html.insert(0, head)

    if i == 0:
        # No ``<html>`` found, add it
        content = h.html(content)

    head1 = presentation.render(h.head, None, None, None)  # The automatically generated ``<head>``

    url = h.request.upath_info.strip('/')
    if url and not head1.xpath('./link[@rel="canonical"]'):
        head1.append(h.head.link(rel='canonical', href=h.request.uscript_name + '/' + url))

    # Merge the attributes and child of the automatically generated ``<head>``
    head.attrib.update(head1.attrib.items())
    head.add_child(head1[:])

    return content



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Import an object from a string reference

The possible reference syntaxes are:

  - ``'python <module>:<object>'`` -- loads an object from a Python module
    (for example: ``'python os.path:isfile'``)
  - ``'<module>:<object>'``  -- same as ``'python <module>:<object>'``
  - ``'file <file>:<object>'`` -- loads an object from a file
    (for example: ``'file /tmp/counter.py:Counter'``)
  - ``'egg <dist>:<app>'`` -- loads the registered application ``<app>``
    from the ``<dist>`` distribution
    (for example: ``'egg nagare:admin'`` or ``'egg nagare.examples:wiki'``)
  - ``'app <app>'`` -- load the registered application ``<app>``
    (for example: ``'app examples'``)
"""

import sys
import os

import pkg_resources


def load_entry_point(app, entry_point):
    """Load an object registered under an entry point

    In:
      - ``app`` -- name of the object
      - ``entry_point`` -- name of the entry_point

    Return:
      - (the object, the distribution of the object)
    """
    apps = dict([(entry.name, entry) for entry in pkg_resources.iter_entry_points(entry_point)])
    entry_point = apps[app]

    return entry_point.load(), entry_point.dist


def load_app(app, _):
    """Load a registered application

    In:
      - ``app`` -- name of the application
      - ``_`` -- *unused**

    Return:
      - (the application, the distribution of the application)
    """
    return load_entry_point(app, 'nagare.applications')


def load_egg(dist, app):
    """Load a registered application of a distribution

    In:
      - ``dist`` -- name of the distribution
      - ``app`` -- name of the application

    Return:
      - (the application, the distribution of the application)
    """
    dist = pkg_resources.get_distribution(dist)
    return dist.get_entry_info('nagare.applications', app).load(), dist


def load_file(filename, app):
    """Load an object from a file

    In:
      - ``filename`` -- name of the file
      - ``app`` -- name of the object to load

    Return:
      - (the object, None)
    """
    dirname = os.path.abspath(os.path.dirname(filename))
    if dirname not in sys.path:
        sys.path.insert(0, dirname)

    name = os.path.splitext(os.path.basename(filename))[0]
    return load_module(name, app)


def load_module(module, app):
    """Load an object from a Python module

    In:
      - ``module`` -- name of the module
      - ``app`` -- name of the object to load

    Return:
      - (the object, None)
    """
    r = __import__(module, fromlist=('',))

    if app is not None:
        r = getattr(r, app)

    return r, None


loaders = {
    '': load_module,
    'python': load_module,
    'egg': load_egg,
    'file': load_file,
    'app': load_app
}


def load_object(reference):
    """Load an object from a reference

    In:
      - ``reference`` -- reference as a string

    Return:
      - a tuple (object loaded, distribution where this object is located or ``None``)
    """
    if ' ' in reference:
        scheme, reference = reference.split(' ', 1)
    else:
        scheme = ''

    if ':' in reference:
        (reference, o) = reference.split(':', 1)
    else:
        o = None

    return loaders[scheme](reference, o)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""Asynchronous update objects and Python to javascript transcoder"""

import types
import inspect
import compiler
import cStringIO
import re

import peak.rules
import pyjs

from nagare import presentation, namespaces, serializer, security, partial

YUI_INTERNAL_PREFIX = '/static/nagare/yui/build'
YUI_EXTERNAL_PREFIX = 'http://yui.yahooapis.com/2.9.0/build'

# ---------------------------------------------------------------------------

YUI_PREFIX = YUI_INTERNAL_PREFIX


class ViewToJs(object):
    def __init__(self, js, id, renderer, output):
        """Wrap a view into a javascript code

        In:
          - ``js`` -- name of the function to call, on the client, to change the
            DOM element
          - ``id`` -- id of the DOM element to replace on the client
          - ``renderer`` -- the current renderer
          - ``output`` -- the view (or ``None``)
        """
        self.js = js
        self.id = id
        self.renderer = renderer
        self.output = output


def serialize_body(view_to_js, content_type, doctype):
    """Wrap a view body into a javascript code

    In:
      - ``view_to_js`` -- the view
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - Javascript to evaluate on the client
    """
    # Get the HTML or XHTML of the view
    body = serializer.serialize(view_to_js.output, content_type, doctype, False)[1]

    # Wrap it into a javascript code
    return "%s('%s', %s)" % (view_to_js.js, view_to_js.id.encode('utf-8'), py2js(body, view_to_js.renderer))


@peak.rules.when(serializer.serialize, (ViewToJs,))
def serialize(self, content_type, doctype, declaration):
    """Wrap a view into a javascript code

    In:
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple ('text/plain', Javascript to evaluate on the client)
    """
    if self.output is None:
        return 'text/plain', ''

    # Get the javascript for the header
    head = presentation.render(self.renderer.head, self.renderer, None, None)

    # Wrap the body and the header into a javascript code
    return 'text/plain', serialize_body(self, content_type, doctype) + '; ' + head


def javascript_dependencies(renderer):
    head = renderer.head

    head.javascript_url(YUI_PREFIX + '/yahoo/yahoo-min.js')
    head.javascript_url(YUI_PREFIX + '/event/event-min.js')
    head.javascript_url(YUI_PREFIX + '/connection/connection-min.js')
    head.javascript_url(YUI_PREFIX + '/get/get-min.js')

    head.javascript('_nagare_content_type_', 'NAGARE_CONTENT_TYPE="%s"' % ('application/xhtml+xml' if renderer.response.xml_output else 'text/html'))


class Update(object):
    """Asynchronous updater object

    Send a XHR request that can do an action, render a component and finally
    update the HTML with the rendered view
    """
    @classmethod
    def no_action(cls, *args, **kw):
        return None

    @partial.max_number_of_args(4)
    def __init__(self, render='', action=None, component_to_update=None, with_request=False, permissions=None, subject=None, args=(), **kw):
        """Initialisation

        In:
          - ``render`` -- can be:

            - a string -- name of the view that will be rendered on the current
              component
            - a function -- this function will be called and the result sent
              back to the client

          - ``action`` -- action to call
          - ``args``, ``kw`` -- ``action`` parameters

          - ``component_to_update`` -- can be:

            - ``None`` -- the current view will be updated
            - a string -- the DOM id of the HTML to update
            - a tag -- this tag will be updated

          - ``with_request`` -- will the request and response objects be passed to the action?

          - ``permissions`` -- permissions needed to execute the action
          - ``subject`` -- subject to test the permissions on
        """
        self.render = render
        self.with_request = with_request
        self.component_to_update = component_to_update

        # Wrap the ``action`` into a wrapper that will check the user permissions
        action = security.wrapper(action or self.no_action, permissions, subject)
        self.action = partial.Partial(action, *args, **kw)

    @classmethod
    def _generate_response(cls, render, args, js, component_to_update, r):
        """Wrap the rendering of a component into a JS statement

        In:
          - ``render`` -- rendering function to call
          - ``args`` -- ``render`` parameters
          - ``js`` -- JS function to wrap the rendering into
          - ``component_to_update`` -- id of the DOM element to update on the client
          - ``r`` -- renderer

        Return:
          - the JS statement
        """
        return ViewToJs(js, component_to_update, r, render(r, *args))

    def _generate_render(self, renderer):
        """Generate the rendering function

        In:
          - ``renderer`` -- the current renderer

        Return:
          - the rendering function
        """
        request = renderer.request

        if request and not request.is_xhr and ('_a' not in request.params):
            javascript_dependencies(renderer)
            renderer.head.javascript_url('/static/nagare/ajax.js')

        js = 'nagare_updateNode'
        component_to_update = self.component_to_update
        if component_to_update is None:
            async_root = renderer.get_async_root()

            # Remember to wrap the root view into a ``<div>``
            component_to_update = async_root.id
            async_root.wrapper_to_generate = True
            js = 'nagare_replaceNode'

        # Get the ``id`` attribute of the target element or, else, generate one
        if isinstance(component_to_update, namespaces.xml._Tag):
            id_ = component_to_update.get('id')
            if id_ is None:
                id_ = renderer.generate_id('id')
                component_to_update.set('id', id_)
            component_to_update = id_

        render = self.render
        model = ()
        if not callable(render):
            model = (render if render != '' else async_root.model,)
            render = renderer.get_async_root().component.render

        return partial.Partial(self._generate_response, render, model, js, component_to_update)

    def _generate_replace(self, priority, renderer):
        """Register the action on the server and generate the javascript to
        the client

        In:
          - ``priority`` -- type of the action (see ``callbacks.py``)
          - ``renderer`` -- the current renderer

        Return:
          - a tuple:

            - id of the action registered on the server
            - name of the javascript function to call on the client
            - id of the tag to update on the client
        """
        return renderer.register_callback(priority, self.action, self.with_request, self._generate_render(renderer))

    def generate_action(self, priority, renderer):
        """Generate the javascript action

        In:
          - ``priority`` -- type of the action (see ``callbacks.py``)
          - ``renderer`` -- the current renderer

        Return:
          - the javascript code
        """
        if priority == 1:
            return 'nagare_getAndEval("%s;_a;%s="+get_field_value(this))' % (renderer.add_sessionid_in_url(sep=';'), self._generate_replace(1, renderer))
        elif priority == 4:
            return 'return nagare_postAndEval(this.form, "%s")' % self._generate_replace(4, renderer)
        elif priority == 41:
            return 'return nagare_getAndEval("%s;_a;%s")' % (renderer.add_sessionid_in_url(sep=';'), self._generate_replace(4, renderer))
        elif priority == 2:
            return 'nagare_getAndEval("%s;_a;%s")' % (renderer.add_sessionid_in_url(sep=';'), self._generate_replace(2, renderer))
        elif priority == 5:
            r = 'return nagare_imageInputSubmit(event, this, "%s")' % self._generate_replace(5, renderer)
            renderer.head.javascript_url(YUI_PREFIX + '/dom/dom-min.js')
            return r
        else:
            return ''


# ---------------------------------------------------------------------------

class ViewsToJs(list):
    """A list of ``ViewToJS`` objects
    """
    pass


@peak.rules.when(serializer.serialize, (ViewsToJs,))
def serialize_views(self, content_type, doctype, declaration):
    """Wrap a view into a javascript code

    In:
      - ``content_type`` -- the rendered content type
      - ``doctype`` -- the (optional) doctype
      - ``declaration`` -- is the XML declaration to be outputed?

    Return:
      - a tuple ('text/plain', Javascript to evaluate on the client)
    """
    bodies = [serialize_body(view_to_js, content_type, doctype) for view_to_js in self if view_to_js.output is not None]

    if not bodies:
        return 'text/plain', ''

    # Get the javascript for the header
    head = presentation.render(self[0].renderer.head, self[0].renderer, None, None)

    return 'text/plain', '; '.join(bodies) + '; ' + head


class Updates(Update):
    """A list of ``Update`` objects
    """
    @partial.max_number_of_args(1)
    def __init__(self, args, action=None, with_request=False, permissions=None, subject=None, **kw):
        """Initialization

        In:
          - ``args`` -- the list of ``Update`` objects
          - ``action`` -- global action to execute
          - ``kw`` -- ``action`` parameters
          - ``with_request`` -- will the request and response objects be passed to the global action?
          - ``permissions`` -- permissions needed to execute the global action
          - ``subject`` -- subject to test the permissions on
        """
        self._updates = args
        self._with_request = with_request
        self._action = action or self.no_action

        super(Updates, self).__init__(action=self.action, with_request=True, permissions=permissions, subject=subject, **kw)

    def action(self, request, response, *args, **kw):
        """Execute all the actions of the ``Update`` objects
        """
        if self._with_request:
            self._action(request, response, *args, **kw)
        else:
            self._action(*args, **kw)

        for update in self._updates:
            if update.with_request:
                update.action(request, response, *args, **kw)
            else:
                update.action(*args, **kw)

    @classmethod
    def _generate_response(cls, r, renders):
        return ViewsToJs(render(r) for render in renders)

    def _generate_render(self, renderer):
        """Generate the rendering function

        In:
          - ``renderer`` -- the current renderer

        Return:
          - the rendering function
        """
        renders = [update._generate_render(renderer) for update in self._updates]
        return partial.Partial(self._generate_response, renders=renders)


# ---------------------------------------------------------------------------

def str2js(src, namespace):
    """Translate a string with Python code to javascript

    In:
      - ``src`` -- the Python code
      - ``namespace`` -- the namespace of this code

    Return:
      - the javascript code
    """
    output = cStringIO.StringIO()
    pyjs.Translator(namespace, compiler.parse(src), output)

    return output.getvalue()


class JS(object):
    """Transcode a Python function or module to javascript code
    """
    def __init__(self, o):
        """Transcode a Python function or module

        In:
          - ``o`` -- Python function or module to transcode
        """
        if hasattr(o, '_js_name'):
            # Already transcoded
            self.name = o._js_name
            self.javascript = o._js_code
            return

        src = inspect.getsource(o)

        classname = ''
        if isinstance(o, types.MethodType):
            classname = o.im_class.__name__
            o = o.im_func

        if isinstance(o, types.FunctionType):
            if o.__module__ == '__main__':
                self.name = module = classname
            else:
                module = o.__module__.replace('.', '_')
                if classname:
                    module += ('_' + classname)
                self.name = module + '_'

            self.name += o.func_name

            indent = inspect.indentsize(src)
            src = ''.join([s.expandtabs()[indent:] for s in src.splitlines(True)])
        else:
            if o.__name__ == '__main__':
                self.name = module = ''
            else:
                self.name = module = o.__name__.replace('.', '_')

        self.javascript = str2js(src, module)

        # Keep the transcoded javascript
        o._js_name = self.name
        o._js_code = self.javascript

    def generate_action(self, priority, renderer):
        """Include the transcoded javascript into ``<head>``

        In:
           - ``priority`` -- *not used*
           - ``renderer`` -- the current renderer

        Return:
          - javascript fragment of the call to the transcoded function
        """
        renderer.head.javascript_url('/static/nagare/pyjslib.js')
        renderer.head.javascript(self.name, self.javascript)

        return self.name + '(this)'


javascript = JS


# ---------------------------------------------------------------------------

@peak.rules.abstract
def py2js(value, h=None):
    """No default transcodage
    """
    pass


@peak.rules.when(py2js, (types.NoneType,))
def py2js(value, h):
    """Generic method to transcode ``None``

    In:
      - ``value`` -- ``None``
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return 'null'


@peak.rules.when(py2js, (bool,))
def py2js(value, h):
    """Generic method to transcode a boolean

    In:
      - ``value`` -- a boolean
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return str(value).lower()


@peak.rules.when(py2js, (int,))
def py2js(value, h):
    """Generic method to transcode an integer

    In:
      - ``value`` -- an integer
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return str(value)


@peak.rules.when(py2js, (long,))
def py2js(value, h):
    """Generic method to transcode a long integer

    In:
      - ``value`` -- a long integer
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return str(value)


@peak.rules.when(py2js, (float,))
def py2js(value, h):
    """Generic method to transcode a float

    In:
      - ``value`` -- a float
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return str(value)


@peak.rules.when(py2js, (dict,))
def py2js(d, h):
    """Generic method to transcode a dictionary

    In:
      - ``value`` -- a dictionary
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return '{' + ', '.join(['%s : %s' % (py2js(name, h), py2js(value, h)) for (name, value) in d.items()]) + '}'


@peak.rules.when(py2js, ((types.FunctionType, types.MethodType),))
def py2js(value, h):
    """Generic method to transcode a function

    In:
      - ``value`` -- a fonction
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return py2js(JS(value), h)


@peak.rules.when(py2js, (types.ModuleType,))
def py2js(value, h):
    """Generic method to transcode a module

    In:
      - ``value`` -- a module
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return py2js(JS(value), h)


@peak.rules.when(py2js, (JS,))
def py2js(value, h):
    """Generic method to transcode an already transcoded function

    In:
      - ``value`` -- a transcoded fonction
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    if h is not None:
        h.head.javascript_url('/static/nagare/pyjslib.js')

    return value.javascript


@peak.rules.when(py2js, (unicode,))
def py2js(value, h):
    """Generic method to transcode an unicode string

    In:
      - ``value`` -- an unicode string
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return py2js(value.encode('utf-8'), h)


not_ascii = re.compile(r'\\x(..)')


@peak.rules.when(py2js, (str,))
def py2js(value, h):
    """Generic method to transcode a string

    In:
      - ``value`` -- a string
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return not_ascii.sub(lambda m: chr(int(m.group(1), 16)), repr(value))


@peak.rules.when(py2js, (list,))
def py2js(l, h):
    """Generic method to transcode a list

    In:
      - ``value`` -- a list
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return '[' + ', '.join([py2js(value, h) for value in l]) + ']'


@peak.rules.when(py2js, (tuple,))
def py2js(value, h):
    """Generic method to transcode a tuple

    In:
      - ``value`` -- a tuple
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return py2js(list(value), h)


class js(str):
    """A javascript code object
    """
    pass


@peak.rules.when(py2js, (js,))
def render(self, h):
    """Generic method to transcode a javascript code

    In:
      - ``value`` -- a javascript code
      - ``h`` -- the current renderer

    Return:
      - transcoded javascript
    """
    return str(self)



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The ``nagare-admin`` executable

The ``run()`` function is called by the ``nagare-admin`` console script, created
by  ``setuptools``
"""

import os
import sys
import optparse

import pkg_resources


# ---------------------------------------------------------------------------

class Command(object):
    """The base class of all the commands"""
    desc = ''

    @classmethod
    def set_options(cls, parser):
        """Define the available options for this command

        Out:
          - ``parser`` -- this ``optparse.OptionParser`` will parse the command line
        """
        pass

    @classmethod
    def run(cls, parser, options, args):
        """Command code

        In:
          - ``parser`` -- this ``optparse.OptionParser`` has parsed the command line
          - ``options`` -- the options read on the command line
          - ``args`` -- arguments left on the command line

        Return:
          - the Unix command status
        """
        return 2

    def parse(self, argv, commands):
        """Parse the command line

        In:
          - ``argv`` -- list of the arguments on the command line
          - ``commands`` -- parent commands of this command

        Return:
          - the Unix command status
        """
        parser = optparse.OptionParser(usage=' '.join(commands))
        self.set_options(parser)

        return self.run(parser, *parser.parse_args(argv))

    def get_description(self, commands):
        """Return the description of the command

        In:
          - ``commands`` -- parent commands of this command

        Return:
          - A list with a single tuple (full name of the command, description of the command)
        """
        return [(' '.join(commands), self.desc)]


class Commands(dict):
    """Command with sub-commands"""

    def add_command(self, names, command):
        """Define a sub-commands

        In:
          - ``names`` -- list of names of the sub-command
          - ``command`` -- the sub-command object
        """
        name = names.pop(0)
        if names:
            # The sub-commands has sub-commands itself.
            # Add a new 'Commands()' node if not already registered
            self.setdefault(name, Commands()).add_command(names, command)
        else:
            self[name] = command

    def _usage(self, commands, error):
        """Usage of this command

        In:
          - ``commands`` -- parent commands of this command
          - ``error`` - optional error message

        Return:
          - yield the usage line per line
        """
        yield 'Usage: ' + ' '.join(commands + ('<command>',))
        yield ''
        yield 'with <command>:'

        # Return the description of each sub-commands
        commands = list(self.get_description(()))

        l = max(len(name) for name, _ in commands)
        for name, desc in sorted(commands):
            yield ' - %s: %s' % (name.ljust(l), desc)

        if error:
            yield ''
            yield error

    def usage(self, commands, error=None):
        """Print the usage of this command on ``stderr``

        In:
          - ``commands`` -- parent commands of this command
          - ``error`` - optional error message

        Return:
          - **no return**, interpreter exit
        """
        print >>sys.stderr, '\n'.join(self._usage(commands, error))
        sys.exit(2)

    def parse(self, argv, commands):
        """Parse the command line

        In:
          - ``argv`` -- list of the arguments on the command line
          - ``commands`` -- parent commands of this command

        Return:
          - the Unix command status
        """
        if not argv:
            # No sub-command given on the command line
            self.usage(commands)

        name = argv.pop(0)
        if name not in self:
            # The sub-command given on the command line doesn't exist
            self.usage(commands, "error: command '%s' not found" % name)

        # Parse the command line for the sub-command given
        return self[name].parse(argv, commands + (name,))

    def get_description(self, commands):
        """Return the description of the command

        In:
          - ``commands`` -- parent commands of this command

        Return:
          - A list of tuples (full name of the sub-command, description of the sub-command)
        """
        for name, command in self.items():
            for r in command.get_description(commands + (name,)):
                yield r


# ---------------------------------------------------------------------------

def run(entry_point_section='nagare.commands'):
    """Dispatcher for the ``nagare-admin`` commands

    The commands are classes, registered under the ``entry_point_section`` entry point
    """

    # Load all the commands
    commands = Commands()
    for entry_point in pkg_resources.iter_entry_points(entry_point_section):
        try:
            commands.add_command(entry_point.name.split('.'), entry_point.load()())
        except ImportError:
            print "Warning: the command '%s' can't be imported" % entry_point.name
            raise

    return commands.parse(sys.argv[1:], (os.path.basename(sys.argv[0]),))



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

from peak.rules import when
import webob

from nagare import security


# ---------------------------------------------------------------------------

# The application can used anything for the permission objects
# So the following pre-defined permissions are optional helpers

class Permission(object):
    """Base class of all the permissions
    """
    pass


class Private(Permission):
    """To define the ``private`` permission singleton

    Nobody has access to objects protected with this permission
    """
    pass


class Public(Permission):
    """To define the ``public`` permission singleton

    Every body has access to objects protected with this permission
    """
    pass


# The singleton permissions
private = Private()
public = Public()


# ---------------------------------------------------------------------------

class Denial(BaseException):
    """Type of the objects return when an access is denied

    In a boolean context, it is evaluated to ``False``
    """
    def __init__(self, message='Access forbidden'):
        """Initialisation

        In:
          - ``message`` -- denial description
        """
        super(Denial, self).__init__(message)

    def __nonzero__(self):
        """Evaluated to ``False`` in a boolean context
        """
        return False

    def __str__(self):
        return 'security.Denial(%s)' % str(self.args[0])


# --------------------------------------------------------------

class User(object):
    """Base class for the user objects
    """
    def __init__(self, id=None, *args):
        self.id = id
        self.credentials = args
        self.expired = False

        self._previous_user = None

    def set_id(self, id, *args):
        self.id = id
        self.credentials = args

    def get_id(self):
        return (self.id,) + self.credentials

    def __enter__(self):
        """Push this user to the stack
        """
        self._previous_user = security.get_user()
        security.set_user(self)

    def __exit__(self, *args, **kw):
        """Pop this user from the stack
        """
        security.set_user(self._previous_user)


# ---------------------------------------------------------------------------

class Rules(object):
    """Pre-defined security rules

    A rule is an implementation of the ``security.common.Rules.has_permission()``
    generic method.
    """
    def __init__(self):
        self._previous_rules = None

    def has_permission(self, user, perm, subject):
        """The ``has_permission()`` generic method
        and default implementation: by default all accesses are denied

        In:
          - ``user`` -- user to check the permission for
          - ``perm`` -- permission(s) to check
          - ``subject`` -- object to check the permission on

        Return:
          - True if the access is granted
          - Else a ``security.common.denial`` object
        """
        return Denial('Access denied')

    @when(has_permission, (object, Private))
    def no_access(self, user, perm, subject):
        """Nobody has access to an object protected with the ``private`` permission
        """
        return Denial()

    @when(has_permission, (object, Public))
    def full_access(self, user, perm, subject):
        """Everybody has access to an object protected with the ``public`` permission
        """
        return True

    @when(has_permission, (User, tuple))
    def check_access_tuple(self, user, perms, subject):
        """If several permissions are to be checked, the access must be granted
        for at least one permission
        """
        return any((self.has_permission(user, perm, subject) for perm in perms)) or Denial('Access forbidden')

    @when(has_permission, (User, list))
    def check_access_list(self, user, perms, subject):
        """If several permissions are to be checked, the access must be granted
        for at least one permission
        """
        return self.has_permission(user, tuple(perms), subject)

    @when(has_permission, (User, set))
    def check_access_set(self, user, perms, subject):
        """If several permissions are to be checked, the access must be granted
        for at least one permission
        """
        return self.has_permission(user, tuple(perms), subject)

    def __enter__(self):
        """Push these security rules to the stack
        """
        self._previous_rules = security.get_manager()
        security.set_manager(self)

    def __exit__(self, *args, **kw):
        """Pop these security rules from the stack
        """
        security.set_manager(self._previous_rules)


# ---------------------------------------------------------------------------

class Authentication(object):
    """An ``Authentication`` object identify, authenticate and create the
    user objects

    .. note::
        By definition, the user object ``None`` is the anonymous user
    """
    def create_user(self, request, response):
        """Check the user is valid and create it
        """
        # Retrieve the data associated with the connected user
        (username, ids) = self.get_ids(request, response)

        # Retrieve the authentication data
        password = self.get_password(username)
        if not self.check_password(username, password, **ids):
            # Bad authentication of the user, create an anonymous user
            username = None

        # Create the user object
        user = self._create_user(username)
        if user:
            self.set_user_id(user, username, **ids)
        return user

    def check_password(self, username, password, **kw):
        """Authentication

        In:
          - ``username`` -- the user id
          - ``password`` -- the real password of the user
          - ``kw`` -- other data for the user

        Return:
          - a boolean
        """
        return (username is None) and (password is None)

    # -----------------------------------------------------------------------

    def get_ids(self, request, response):
        """Return the data associated with the connected user

        In:
          - ``request`` -- the web request object
          - ``response`` -- the web response object

        Return:
          - A tuple with the id of the user and a dictionary of its data
        """
        return (None, {})

    def set_user_id(self, user, id, **kw):
        """Set the credentials of the user

        In:
          - ``user`` -- the user
          - ``id`` -- the user id
          - ``**kw`` -- the user credentials
        """
        user.set_id(id)

    def logout(self):
        """Deconnection of the current user
        """
        return None

    def denies(self, detail):
        """Method called when a permission is denied

        In:
          - ``detail`` -- a ``security.common.denial`` object
        """
        raise webob.exc.HTTPForbidden(str(detail))

    def get_password(self, username):
        """Return the real password of the user

        In:
          - ``username`` -- the user id

        Return:
          - the password
        """
        return None

    def _create_user(self, username):
        """The user is validated, create the user object

        In:
          - ``username`` -- the user id

        Return:
          - the user object
        """
        return None

    def end_rendering(self, request, response, session):
        """End of the request processing

        In:
          - ``request`` -- the request object
          - ``response`` -- the response object
          - ``session`` -- the session
        """
        pass



# --
# Copyright (c) 2008-2018 Net-ng.
# All rights reserved.
#
# This software is licensed under the BSD License, as described in
# the file LICENSE.txt, which you should have received as part of
# this distribution.
# --

"""The FastCGI publisher"""

import os
import time
from flup.server import fcgi_fork

from nagare import local
from nagare.publishers import common


class Publisher(common.Publisher, fcgi_fork.WSGIServer):
    """The FastCGI publisher"""

    default_port = 9000  # Default FastCGI port

    # Possible command line options with the default values
    # ------------------------------------------------------

    spec = dict(
        host='string(default=None)', port='integer(default=9000)',
        socket='string(default="")', umask='integer(default=None)',
        multiplexed='boolean(default=None)', minSpare='integer(default=None)',
        maxSpare='integer(default=None)', maxChildren='integer(default=None)',
        maxRequests='integer(default=None)',
    )

    def __init__(self):
        """Initialization
        """
        super(Publisher, self).__init__()

        local.worker = local.Process()
        local.request = local.Process()

    def _sanitizeEnv(self, environ):
        super(Publisher, self)._sanitizeEnv(environ)
        environ['SCRIPT_NAME'] = ''

    def _child(self, sock, parent):
        """A new process is created

        In:
          - ``sock`` -- the inherited socket
          - ``parent`` -- the parent process
        """
        # Call each time the ``on_new_process()`` method
        self.on_new_process()
        super(Publisher, self)._child(sock, parent)

    def register_static(self, name, get_file):
        """Register a WSGI application to serve static contents

        In:
          - ``name`` -- the URL of the contents will be prefix by ``/static/<name>/``
          - ``get_file`` -- function that will received the URL of the static content
            and will return its filename

        Return:
          - URL prefix (``/static/<name>/``)

        .. warning::
           **Dummy method**

           The front HTTP server must be configurated to serve the static contents

           (you can use the administrative command ``nagare-admin create-rules`` to generate
           the rewrite rules if you are using the lighttpd or nginx servers)
        """
        return '/static/' + name + '/'

    def _validate_conf(self, filename, conf, error):
        """Validate the configuration read from the publisher configuration file

       In:
         - ``filename`` -- the path to the configuration file
         - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
         - ``error`` -- the function to call in case of configuration errors

       Return:
         - the tuple:

           - hostname to listen to
           - port to listen to
           - unix socket to listen to
           - conf object
       """
        (host, port, conf) = super(Publisher, self)._validate_conf(filename, conf, error)
        return host, port, conf.pop('socket'), conf

    def serve(self, filename, conf, error):
        """Run the publisher

        In:
          - ``filename`` -- the path to the configuration file
          - ``conf`` -- the ``ConfigObj`` object, created from the configuration file
          - ``error`` -- the function to call in case of configuration errors
        """
        (host, port, socket, conf) = self._validate_conf(filename, conf, error)

        if not socket:
            bind_address = (host, port)
            banner = 'fastcgi://%s:%d'
        else:
            bind_address = os.path.normpath(socket)
            banner = 'unix://%s'

        fcgi_fork.WSGIServer.__init__(self, self.urls, bindAddress=bind_address, debug=False, **conf)
        print time.strftime('%x %X -', time.localtime()), 'serving on', banner % bind_address
        self.run()
